import numpy as np
import networkx as nx
from .shannon import local_differential_entropy

from .utils import COV_NULL, make_powerset
from ..lattices import LATTICE_2, LATTICE_3, LATTICE_4
from numpy.typing import NDArray

Atom = tuple[tuple[int, ...], ...]
BOTTOM_2: Atom = ((0,), (1,))
BOTTOM_3: Atom = ((0,), (1,), (2,))
BOTTOM_4: Atom = ((0,), (1,), (2,), (3,))

PATHS_2: dict = nx.shortest_paths.shortest_path_length(
    LATTICE_2, source=None, target=BOTTOM_2
)
LAYERS_2: dict[int, int] = {
    val: {key for key in PATHS_2.keys() if PATHS_2[key] == val}
    for val in set(PATHS_2.values())
}

PATHS_3: dict[int, int] = nx.shortest_paths.shortest_path_length(
    LATTICE_3, source=None, target=BOTTOM_3
)
LAYERS_3: dict[int, int] = {
    val: {key for key in PATHS_3.keys() if PATHS_3[key] == val}
    for val in set(PATHS_3.values())
}

PATHS_4: dict = nx.shortest_paths.shortest_path_length(
    LATTICE_4, source=None, target=BOTTOM_4
)
LAYERS_4: dict = {
    val: {key for key in PATHS_4.keys() if PATHS_4[key] == val}
    for val in set(PATHS_4.values())
}
def unpack_atom(atom: tuple) -> set:
    """
    A utitlity function to unpack tuples.
    """
    varset: set = set()
    for s in {*atom}:
        varset.update(set(s))

    return varset


def local_precompute_sources(data: NDArray[np.floating], cov: NDArray[np.floating] = COV_NULL):
    """
    A utility function that pre-computes the local entropies
    of every subset of the data. This speeds up the PID/GID/PED
    by orders of magnitude.

    Parameters
    ----------
    data : NDArray[np.floating]
        The data in channels x samples format.
    cov : NDArray[np.floating], optional
        The covariance matrix of the data.
        The default is COV_NULL.

    Returns
    -------
    dict
        A dictionary of the local entropies for every source.

    """
    if cov[0][0] == -1:
        cov = np.cov(data, ddof=0.0)

    N: int = data.shape[0]
    joint: tuple = tuple(i for i in range(N))
    sources: list = list(make_powerset(joint))
    sources.remove(())

    return {
        source: local_differential_entropy(data[source, :], cov[source, :][:, source])
        for source in sources
    }


def hmin_differential_redundancy(
    atom: tuple, sources: dict, data: NDArray[np.floating], cov: NDArray[np.floating] = COV_NULL
) -> NDArray[np.floating]:
    """
    For a collection of sources :math:`\\alpha=\\{a_1, a_2, \\ldots, a_k\\}`, computes
    the redundnat entropy shared by all sources as:

    :math:`h_{\\cap}^{min}(\\alpha) = \\min_{i}(a_{i})`

    See:
        Finn, C., & Lizier, J. T. (2020).
        Generalised Measures of Multivariate Information Content.
        Entropy, 22(2), Article 2.
        https://doi.org/10.3390/e22020216

    Parameters
    ----------
    atom : tuple
        The partial information atom. In the form :math:`((a_1,),(a_2,)\\ldots)`.
    sources : dict
        The pre-computed collection of sources.
    data : NDArray[np.floating]
        The data, assumed to be in sources x samples format.

    Returns
    -------
    NDArray[np.floating]
        The local redundancies for each sample.


    """
    h_min: NDArray[np.floating] = np.repeat(np.inf, data.shape[1])
    source: tuple = tuple()

    for source in atom:
        # i+ = min(h(source))
        h_min = np.minimum(h_min, sources[source])

    return h_min


def imin_differential_redundancy(
    atom: tuple,
    sources: dict,
    inputs: tuple,
    target: tuple,
    data: NDArray[np.floating],
    cov: NDArray[np.floating] = COV_NULL,
) -> NDArray[np.floating]:
    """
    Computes the differential redundnacy between a partial information atom :math:`\\alpha=\\{a_i,\\ldots,a_k\\}` and a (potentially multivariate) target. Uses a Gaussian estimator for the local entropies.

    :math:`i_{\\cap}^{min}(\\alpha;t) = \\min_{i}h(a_i) - \\min_{i}h(a_i|t)`

    This is NOT the I_min from Williams and Beer - it is the local redundancy function from Finn and Lizier.

    See:
        Finn, C., & Lizier, J. T. (2018).
        Pointwise Partial Information Decomposition Using
        the Specificity and Ambiguity Lattices.
        Entropy, 20(4), Article 4.
        https://doi.org/10.3390/e20040297

    Parameters
    ----------
    atom : tuple
        The partial information atom. In the form :math:`((a_1,),(a_2,)\\ldots)`.
    sources : dict
        The pre-computed collection of sources.
    inputs : tuple
        The indicies of the inputs - one index per element of the tuple.
    target : tuple
        The (potentially multivariate) indices of the collective target.
    data : NDArray[np.floating]
        The data, assumed to be in channels x samples format.

    Returns
    -------
    NDArray[np.floating]
        The local differnetial redundancy for each frame.

    """
    assert data.shape[0] == (len(inputs) + len(target))
    if cov[0][0] == -1:
        cov = np.cov(data, ddof=0)

    i_plus: NDArray[np.floating] = np.repeat(np.inf, data.shape[1])
    i_minus: NDArray[np.floating] = np.repeat(np.inf, data.shape[1])

    h_target: NDArray[np.floating] = sources[
        target
    ]  # No need to transform it. It is the variable idx.

    source: tuple = tuple()
    for (
        source
    ) in atom:  # All atom idxs need to be transformed into source variable idxs.
        # Atom idx to source variable idx conversion.
        # Most of the time this is redundnat, but in case someone tries to call
        # the redundancy function directly from a big time series array, this
        # conversion is important for safety.
        source_inputs: tuple = tuple(inputs[x] for x in source)
        joint_inputs: tuple = tuple(sorted(source_inputs + target))

        h_joint: NDArray[np.floating] = sources[joint_inputs]

        h_conditional: NDArray[np.floating] = h_joint - h_target

        # i+ = min(h(source))
        i_plus = np.minimum(i_plus, sources[source_inputs])
        # i- = min(h(source|target))
        i_minus = np.minimum(i_minus, h_conditional)

    i_min: NDArray[np.floating] = i_plus - i_minus

    return i_min


def mobius_inversion(
    decomposition: str,
    data: NDArray[np.floating],
    inputs: tuple,
    target: tuple = (None,),
    cov: NDArray[np.floating] = COV_NULL,
) -> tuple[dict, dict]:
    """
    Computes the Mobius inversion on a lattice, given a redundancy function.

    See:
        Williams, P. L., & Beer, R. D. (2010).
        Nonnegative Decomposition of Multivariate Information.
        arXiv:1004.2515 [Math-Ph, Physics:Physics, q-Bio].
        http://arxiv.org/abs/1004.2515

    Parameters
    ----------
    decomposition : str
        Whehter to do a PID or PED. Options: "pid", "ped".
    data : NDArray[np.floating]
        The data, assumed to be in channels x samples format.
    inputs : tuple
        The variables to be decomposed.
    target : tuple, optional
        If doing a PID, the indices of the target.
        If doing a PED, leave false.
        The default is (None,).
    cov : NDArray[np.floating], optional
        The covariance matrix that defines the multivairate distribution.
        If left empty, it is computed from the data. The default is COV_NULL.

    Returns
    -------
    ptw : dict
        The pointwise values of each atom in the decomposition for each sample.
    avg : dict
        The expected values of each atom in the decomposition.

    """

    decomposition_lower: str = decomposition.lower()
    assert decomposition_lower in {
        "pid",
        "ped",
    }, "You must specify a decomposition: PID or PED."

    if decomposition_lower == "pid":
        assert target != (None,), "You must specify a target."

    if cov[0][0] == -1:
        cov = np.cov(data, ddof=0)

    if len(inputs) == 2:
        lattice, bottom, layers = LATTICE_2, BOTTOM_2, LAYERS_2
    elif len(inputs) == 3:  #    |
        lattice, bottom, layers = LATTICE_3, BOTTOM_3, LAYERS_3
    else:  #                      |
        lattice, bottom, layers = LATTICE_4, BOTTOM_4, LAYERS_4

    # Pre-computing all the local h_{\partial}(source) values
    # this saves a lot of time.

    sources: dict = local_precompute_sources(data=data, cov=cov)

    total_str: str = ""
    partial_str: str = ""

    if decomposition_lower == "pid":
        total_str = "total_information"
        partial_str = "pi"
    elif decomposition_lower == "ped":
        total_str = "total_entropy"
        partial_str = "pe"

    for layer in layers:
        for atom in layers[layer]:
            if decomposition_lower == "pid":
                lattice.nodes[atom][total_str] = imin_differential_redundancy(
                    atom, sources, inputs, target, data, cov
                )
            elif decomposition_lower == "ped":
                lattice.nodes[atom][total_str] = hmin_differential_redundancy(
                    atom, sources, data, cov
                )

            if atom == bottom:
                lattice.nodes[atom][partial_str] = lattice.nodes[atom][total_str]
            else:
                lattice.nodes[atom][partial_str] = lattice.nodes[atom][total_str] - sum(
                    [
                        lattice.nodes[d][partial_str]
                        for d in lattice.nodes[atom]["descendants"]
                    ]
                )

    ptw: dict = {node: lattice.nodes[node][partial_str] for node in lattice.nodes}
    avg: dict = {
        node: lattice.nodes[node][partial_str].mean() for node in lattice.nodes
    }

    return (ptw, avg)

def partial_information_decomposition(
    inputs: tuple[int, ...],
    target: tuple[int, ...],
    data: NDArray[np.floating],
    cov: NDArray[np.floating] = COV_NULL,
) -> tuple[dict, dict]:
    """
    The pointwise and average partial information decomposition
    using the Gaussian imin function. See:
    
    See:
        Finn, C., & Lizier, J. T. (2018).
        Pointwise Partial Information Decomposition Using the Specificity and Ambiguity Lattices.
        Entropy, 20(4), Article 4.
        https://doi.org/10.3390/e20040297

    Parameters
    ----------
    inputs : tuple[int, ...]
        The indices of the input variables.
    target : tuple[int, ...]
        The indices of the target variable(s).
    data : NDArray[np.floating]
        The data in channels x time format.
    cov : NDArray[np.floating]
        The covariance matrix of the data.
        The default is COV_NULL.

    Returns
    -------
    ptw : dict
        The pointwise PID for every frame in the data.
    avg : dict
        The average PID for the data

    """

    if cov[0][0] == -1:
        cov = np.cov(data, ddof=0)
    else:
        assert cov.shape[0] == data.shape[0], (
            "The covariance matrix must have the same dimensionality as the data."
        )

    N_inputs: int = len(inputs)
    N_target: int = len(target)
    joint: tuple = inputs + target

    ptw, avg = mobius_inversion(
        decomposition="pid",
        data=np.vstack((data[inputs, :], data[target, :])),
        inputs=tuple(i for i in range(N_inputs)),
        target=tuple(i + N_inputs for i in range(N_target)),
        cov=cov[np.ix_(joint, joint)],
    )

    return ptw, avg


def partial_entropy_decomposition(
    inputs: tuple[int, ...],
    data: NDArray[np.floating],
    cov: NDArray[np.floating] = COV_NULL,
) -> tuple[dict, dict]:
    """
    Computes the partial entropy decomposition of a joint distribution
    with up to four elements. Uses the Gaussian hmin estimator. See:

        Finn, C., & Lizier, J. T. (2020).
        Generalised Measures of Multivariate Information Content.
        Entropy, 22(2), Article 2.
        https://doi.org/10.3390/e22020216


    Parameters
    ----------
     inputs : tuple[int, ...]
        The indices of the elements to analyze.
    data : NDArray[np.floating]
        The numpy array, assumed to be in channels x time format.
    cov : NDArray[np.floating], optional
        The covariance matrix. If none is not provided, it is computed
        from the data directly.

    Returns
    -------
    ptw : dict
        The dictionary of local values for each partial entropy atom.
        The local values are represented by a numpy array of the same
        length as the data array.
    avg : dict
        The dictionary of expected values for each partial entropy atom.

    """
    if cov[0][0] == -1:
        cov = np.cov(data, ddof=0)
    else:
        assert cov.shape[0] == data.shape[0], (
            "The covariance matrix must have the same dimensionality as the data."
        )

    N_inputs: int = len(inputs)

    ptw, avg = mobius_inversion(
        decomposition="ped",
        data=data[inputs, :],
        inputs=tuple(i for i in range(N_inputs)),
        cov=cov[np.ix_(inputs, inputs)],
    )

    return ptw, avg


def generalized_information_decomposition(
    inputs: tuple[int, ...],
    data: NDArray[np.floating],
    cov_posterior: NDArray[np.floating],
    cov_prior: NDArray[np.floating],
) -> tuple[dict, dict]:
    """
    Computes the generalized information decomposition from Varley et al.
    The GID is a decomposition of the Kullback-Leibler divergence of a
    posterior distribution from a prior distribution.

    The available redundancy function is the Gaussian hmin.

    See:
        Varley, T. F. (2024).
        Generalized decomposition of multivariate information.
        PLOS ONE, 19(2), e0297128.
        https://doi.org/10.1371/journal.pone.0297128

    Parameters
    ----------
     inputs : tuple[int, ...]
        The indices of the elements to analyze.
    data : NDArray[np.floating]
        The numpy array, assumed to be in channels x time format.
    cov_posterior : NDArray[np.floating]
        The covariance matrix that defines the prior distribution.
    cov_prior : NDArray[np.floating]
        The covariance matrix that defines the posterior distribution.

    Returns
    -------
    ptw : dict
        The dictionary of local values for each partial entropy atom.
        The local values are represented by a numpy array of the same
        length as the data array.
    avg : dict
        The dictionary of expected values for each partial entropy atom.

    """

    N_inputs: int = len(inputs)

    ptw_prior, _ = mobius_inversion(
        decomposition="ped",
        data=data[inputs, :],
        inputs=tuple(i for i in range(N_inputs)),
        cov=cov_prior[np.ix_(inputs, inputs)],
    )

    ptw_posterior, _ = mobius_inversion(
        decomposition="ped",
        data=data[inputs, :],
        inputs=tuple(i for i in range(N_inputs)),
        cov=cov_posterior[np.ix_(inputs, inputs)],
    )

    ptw: dict = {key: ptw_prior[key] - ptw_posterior[key] for key in ptw_prior.keys()}
    avg: dict = {key: ptw[key].mean() for key in ptw.keys()}

    return ptw, avg


def representational_complexity(avg: dict, comparator=min) -> float:
    """
    Computes the representational complexity of a given partial information or entropy lattice.
    The representational complexity is a measure of how
    much partial information atoms of a given degree of synergy
    contribute to the overall mutual information or entropy.

    See:
        Ehrlich, D. A., Schneider, A. C., Priesemann, V., Wibral, M., & Makkeh, A. (2023).
        A Measure of the Complexity of Neural Representations
        based on Partial Information Decomposition.
        Transactions on Machine Learning Research.
        https://openreview.net/forum?id=R8TU3pfzFr



    Parameters
    ----------
    avg : dict
        The dictionary of partial information/entropy atoms.
        Returned from any of the above functions.
    comparator : function, optional
        Whether to consider the minimum complexity of an atom.
        or the maximum complexity of an atom.
        Options are: min, max, np.min, np.max.
        The default is min, following the original work
        by Ehrlich et al.,.

    Returns
    -------
    float
        The representational complexity.

    """

    assert comparator in (min, max, np.min, np.max), "The comparator must be min or max"

    rc: float = 0.0

    for atom in avg.keys():
        rc += avg[atom] * comparator(len(source) for source in atom)

    return rc / sum(avg.values())



from .shannon import (
    differential_entropy,
    mutual_information,
    conditional_mutual_information
)

from .multivariate_mi import (
     total_correlation,
     dual_total_correlation,
     s_information,
     o_information
)

__all__ = [
    "differential_entropy",
    "mutual_information",
    "conditional_mutual_information",
     "total_correlation",
     "dual_total_correlation",
     "s_information",
     "o_information",
]



from .shannon import (
        shannon_entropy, 
        conditional_entropy, 
        mutual_information
        )

__all__ = [
        "shannon_entropy",
        "conditional_entropy",
        "mutual_information"
        ]



import numpy as np
import itertools as it
from scipy.special import comb
from .utils import (
    reduce_state,
    get_marginal_distribution,
    marginalize_out,
    get_all_marginal_distributions,
)
from .shannon import kullback_leibler_divergence, shannon_entropy
from .optimization import constrained_maximum_entropy_distributions

binom_lookup = {N: {k: comb(N, k, exact=True) for k in range(N)} for N in range(16)}


def total_correlation(
    joint_distribution: dict[tuple[int, ...], float],
) -> (tuple[dict, float], float):
    """
    Computes the average and pointwise total correlations:

    .. math::

        TC(X) &= D_{KL}(P(X) || \\prod_{i=1}^{N}P(X_i) \\\\
              &= \\sum_{i=1}^{N}H(X_i) - H(X)

    Parameters
    ----------
    joint_distribution : dict[tuple[int, ...]], float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    ptw : dict[tuple, float]
        The pointwise TC .
    avg : float
        The average TC.
    
    References
    ----------
    Watanabe, S. (1960). Information Theoretical Analysis of Multivariate Correlation.
    IBM Journal of Research and Development, 4(1), Article 1.
    https://doi.org/10.1147/rd.41.0066

    Tononi, G., Sporns, O., & Edelman, G. M. (1994). 
    A measure for brain complexity: Relating functional segregation and integration in the nervous system. 
    Proceedings of the National Academy of Sciences, 91(11), Article 11. 
    https://doi.org/10.1073/pnas.91.11.5033
    
    """

    maxent_distribution = constrained_maximum_entropy_distributions(
        joint_distribution, order=1
    )

    ptw, avg = kullback_leibler_divergence(joint_distribution, maxent_distribution)

    return ptw, avg


def delta_k(
    k: int, joint_distribution: dict[tuple[int, ...], float]
) -> (tuple[dict, float], float):
    r"""
    S-information, DTC, and negative O-information can all be written in a general form:

    .. math::

        \Delta^{k}(X) = (N-k)TC(X) - \sum_{i=1}^{N}TC(X^{-i})

    Parameters
    ----------
    k : int
        The scale parameter
    joint_distribution : dict[tuple[int, ...], float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    ptw : dict[tuple[int, ...], float]
        The pointwise :math:`\delta^{k}`.
    avg : float
        The average :math:`\Delta^{k}`.

    References
    ----------
    Varley, T. F., Pope, M., Faskowitz, J., & Sporns, O. (2023).
    Multivariate information theory uncovers synergistic subsystems of the human cerebral cortex.
    Communications Biology, 6(1), Article 1.
    https://doi.org/10.1038/s42003-023-04843-w

    """

    states: list[tuple[int, ...]] = list(joint_distribution.keys())
    N: int = len(states[0])

    ptw_tc: dict
    avg_tc: float

    ptw_tc, avg_tc = total_correlation(joint_distribution)

    avg_whole: float = (N - k) * avg_tc
    avg_sum_parts: float = 0.0

    ptw_whole = {state: ptw_tc[state] * (N - k) for state in states}
    ptw_sum_parts = {state: 0 for state in states}

    for i in range(N):
        residuals: tuple = tuple(j for j in range(N) if j != i)

        reduced_distribution: dict = marginalize_out((i,), joint_distribution)

        ptw_r: dict
        avg_r: dict
        ptw_r, avg_r = total_correlation(reduced_distribution)

        avg_sum_parts += avg_r

        for state in ptw_r:
            full_states = {s for s in states if reduce_state(s, residuals) == state}

            for full_state in full_states:
                ptw_sum_parts[full_state] += ptw_r[state]

    ptw: dict = {state: ptw_whole[state] - ptw_sum_parts[state] for state in states}
    avg: float = avg_whole - avg_sum_parts

    return ptw, avg


def s_information(
    joint_distribution: dict[tuple[int, ...], float],
) -> (tuple[dict, float], float):
    """
    Computes the local and expected S-information for the joint distribution.

    .. math:: 

        \\Sigma(X) &= \\sum_{i=1}^{N}I(X_i;X^{-i}) \\\\
                   &= N\\times TC(X) - \\sum_{i=1}^{N}TC(X^{-i}) \\\\
                   &= TC(X) + DTC(X)
        
    Parameters
    ----------
    joint_distribution : dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    ptw : dict[tuple, float]
        The pointwise S-information .
    avg : float
        The average S-information.

    References
    ----------
    Rosas, F., Mediano, P. A. M., Gastpar, M., & Jensen, H. J. (2019).
    Quantifying High-order Interdependencies via Multivariate Extensions of the Mutual Information.
    Physical Review E, 100(3), Article 3.
    https://doi.org/10.1103/PhysRevE.100.032305

    Varley, T. F., Pope, M., Faskowitz, J., & Sporns, O. (2023).
    Multivariate information theory uncovers synergistic subsystems of the human cerebral cortex.
    Communications Biology, 6(1), Article 1.
    https://doi.org/10.1038/s42003-023-04843-w

    """

    ptw: dict
    avg: float
    ptw, avg = delta_k(k=0, joint_distribution=joint_distribution)

    return ptw, avg


def dual_total_correlation(
    joint_distribution: dict[tuple[int, ...], float],
) -> (tuple[dict, float], float):
    """
    Computes the local and expected dual total correlations for the joint distribution.

    .. math:: 

        DTC(X) &= H(X) - \\sum_{i=1}^{N}H(X_i|X^{-i}) \\\\
               &= (N-1)\\times TC(X) - \\sum_{i=1}^{N}TC(X^{-i})

    Parameters
    ----------
    joint_distribution : dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    ptw : dict[tuple, float]
        The pointwise DTC .
    avg : float
        The average DTC.

    References
    ----------
    Abdallah, S. A., & Plumbley, M. D. (2012).
    A measure of statistical complexity based on predictive information with application to finite spin systems.
    Physics Letters A, 376(4), 275–281.
    https://doi.org/10.1016/j.physleta.2011.10.066

    Rosas, F., Mediano, P. A. M., Gastpar, M., & Jensen, H. J. (2019).
    Quantifying High-order Interdependencies via Multivariate Extensions of the Mutual Information.
    Physical Review E, 100(3), Article 3.
    https://doi.org/10.1103/PhysRevE.100.032305

    """
    ptw: dict
    avg: float
    ptw, avg = delta_k(k=1, joint_distribution=joint_distribution)

    return ptw, avg


def o_information(
    joint_distribution: dict[tuple[int, ...], float],
) -> (tuple[dict, float], float):
    """
    Computes the local and expected O-informations for the joint distribution.
    O-information quantifies the balance between redundancy (positive values) and synergy (negative values) in multivariate information.

    .. math::

        \\Omega(X) &= (2-N)TC(X) + \\sum_{i=1}^{N}TC(X^{-i}) \\\\
                   &= TC(X) - DTC(X)

    Parameters
    ----------
    joint_distribution : dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    ptw : dict[tuple, float]
        The pointwise O-information .
    avg : float
        The average O-information.
    
    References
    ----------
    Rosas, F., Mediano, P. A. M., Gastpar, M., & Jensen, H. J. (2019).
    Quantifying High-order Interdependencies via Multivariate Extensions of the Mutual Information.
    Physical Review E, 100(3), Article 3.
    https://doi.org/10.1103/PhysRevE.100.032305

    Varley, T. F., Pope, M., Faskowitz, J., & Sporns, O. (2023).
    Multivariate information theory uncovers synergistic subsystems of the human cerebral cortex.
    Communications Biology, 6(1), Article 1.
    https://doi.org/10.1038/s42003-023-04843-w

    """

    ptw: dict
    avg: float
    ptw, avg = delta_k(k=2, joint_distribution=joint_distribution)

    return {state: -ptw[state] for state in ptw.keys()}, -avg


def co_information(
    joint_distribution: dict[tuple[int, ...], float],
) -> (tuple[dict, float], float):
    """
    Computes the cO-information, the third generalization of bivariate mutual information. Unlike total correlation and dual total correlation, the cO-information can be negative and is difficult to interpret.

    .. math::
        Co(X) = \\sum_{\\xi\\subseteq X}(-1)^{|\\xi|}H(\\xi)

    Parameters
    ----------
    joint_distribution : dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    ptw : dict[tuple, float]
        The pointwise cO-information.
    avg : float
        The average cO-information.

    References
    ----------
    Bell, A. J. (2003, April).
    The Co-information lattice.
    4th International Symposium on Independent Component Analysis and
    Blind Signal Separation, Nara, Japan.
    https://www.semanticscholar.org/paper/THE-CO-INFORMATION-LATTICE-Bell/25a0cd8d486d5ffd204485685226f189e6eadd4d

    """

    # Get the lattice of marginal distributions.
    marginals: dict = get_all_marginal_distributions(joint_distribution)
    # Convert them to local entropies as a batch.
    h_marginals = {key: shannon_entropy(marginals[key])[0] for key in marginals}

    ptw: dict = {state: 0.0 for state in joint_distribution.keys()}
    avg: float = 0.0

    for state in joint_distribution.keys():
        for source in h_marginals.keys():
            sign = (-1) ** len(source)

            ptw[state] -= sign * h_marginals[source][reduce_state(state, source)]

        avg += joint_distribution[state] * ptw[state]

    return ptw, avg


def tse_complexity(
    joint_distribution: dict[tuple[int, ...], float], num_samples
) -> float:
    """
    The Tononi-Sporns-Edelman neural complexity measure, which provides a measure of the balance between integration and segregation across scales.

    .. math::

        TSE(X) &= \\sum_{k=1}^{\\lfloor N/2\\rfloor} \\bigg\\langle I(X^{k}_j;X^{-k}_j) \\bigg\\rangle_{j} \\\\
               &= \\sum_{k=2}^{N}\\bigg[\\bigg(\\frac{k}{N}\\bigg)TC(X) - \\langle TC(X^{k}_{j}) \\rangle_{j}  \\bigg] 

    Runtimes scale very badly with system size (as it requires brute-forcing) all possible bipartitions of the system. If the system is too large, a sub-sampling approach is taken: at each scale, num_samples are drawn from the space of bipartitions.

    Parameters
    ----------
    joint_distribution : dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.
    num_samples : int
        The number of samples to do for each subset size..

    Returns
    -------
    float
        The TSE complexity. No local complexity is computed. .

    References
    ----------
    Tononi, G., Sporns, O., & Edelman, G. M. (1994).
    A measure for brain complexity: Relating functional segregation and integration in the nervous system.
    Proceedings of the National Academy of Sciences, 91(11), Article 11.
    https://doi.org/10.1073/pnas.91.11.5033

    Varley, T. F., Pope, M., Faskowitz, J., & Sporns, O. (2023).
    Multivariate information theory uncovers synergistic subsystems of the human cerebral cortex.
    Communications Biology, 6(1), Article 1.
    https://doi.org/10.1038/s42003-023-04843-w

    """
    N: int = len(list(joint_distribution.keys())[0])

    tc_whole: float = total_correlation(joint_distribution)[1]

    null_tcs: np.ndarray = np.array(
        [(float(i) / float(N)) * tc_whole for i in range(1, N + 1)]
    )
    exp_tcs: np.ndarray = np.zeros(null_tcs.shape[0])
    exp_tcs[-1] = tc_whole

    for k in range(1, N):  # For each layer of the TSE-leaf.
        binom: int = binom_lookup[N][k]

        if (
            binom > num_samples
        ):  # if N-choose-k is greater than the pre-specified number of samples:
            # Samples is a set to avoid repeats.
            samples: set = {
                tuple(sorted(tuple(np.random.choice(N, size=k, replace=False))))
                for i in range(num_samples)
            }
        else:
            samples: set = set(it.combinations(range(N), k))

        tcs: float = 0.0

        sample: tuple
        for sample in samples:
            marginal: dict = get_marginal_distribution(sample, joint_distribution)
            tcs += total_correlation(marginal)[1]

        exp_tcs[k - 1] = tcs / len(samples)

    return (null_tcs - exp_tcs).sum()


def description_complexity(
    joint_distribution: dict[tuple[int, ...], float],
) -> (tuple[dict, float], float):
    """
    The description complexity was proposed by Tononi and Sporns as a
    heuristic, easy-to-compute approximation of the full TSE-Complexity.
    Later shown by Varley et al., to be directly proportional to
    the dual total correlation.

    .. math::

        C(X) = \\frac{DTC(X)}{N}


    Where :math:`N` is the number of elements in :math:`X`.

    Parameters
    ----------
    joint_distribution : dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    ptw : dict[tuple, float]
        The pointwise description complexity.
    avg : float
        The average description complexity.

    References
    ----------
    Varley, T. F., Pope, M., Faskowitz, J., & Sporns, O. (2023).
    Multivariate information theory uncovers synergistic subsystems of the human cerebral cortex.
    Communications Biology, 6(1), Article 1.
    https://doi.org/10.1038/s42003-023-04843-w

    """

    N = float(len(list(joint_distribution.keys())[0]))

    ptw: dict
    avg: float
    ptw, avg = delta_k(k=1, joint_distribution=joint_distribution)

    avg /= N
    ptw = {key: ptw[key] / N for key in ptw.keys()}

    return ptw, avg


def connected_information(
    joint_distribution: dict[tuple[int, ...], float], maximum_order: int = -1
) -> list[float]:
    """
    Returns the connected information profile from Schneidman et al.,
    which decomposes the total correlation into contributing parts of
    different orders:

    .. math::
        TC(X) = \\sum_{k=2}^{N}TC^{k}(X)

    Where the `k` superscript refers to the maximum-entropy distribution that preserves all marginals of order `k`.

    One of the few measures that can reliably distinguish between the
    JAMES_DYADIC and JAMES_TRIADIC distributions.

    Parameters
    ----------
    joint_distribution : dict[tuple[int, ...], float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.
    maximum_order : int, optional
        The highest order of marginals to sweep.
        The default sweeps all.

    Returns
    -------
    list[float]
        The connected information profile.

    References
    ----------
    Schneidman, E., Still, S., Berry, M. J., & Bialek, W. (2003).
    Network Information and Connected Correlations.
    Physical Review Letters, 91(23), 238701.
    https://doi.org/10.1103/PhysRevLett.91.238701

    """

    N: int = len(list(joint_distribution.keys())[0])
    if maximum_order == -1:
        maximum_order = N
    
    profile: list[float] = []

    for order in range(1, maximum_order + 1):
        maxent: dict[tuple[int, ...], float] = (
            constrained_maximum_entropy_distributions(
                joint_distribution=joint_distribution, order=order
            )
        )
        profile.append(shannon_entropy(maxent)[1])

    profile = [profile[i-1] - profile[i] for i in range(1, len(profile))]

    return profile



import itertools
import numpy as np

from .shannon import conditional_entropy
from .optimization import constrained_maximum_entropy_distributions
from .utils import get_marginal_distribution, reduce_state

from typing import Any

DiscreteDist = dict[tuple[Any, ...], float]
AlphaSynDist = dict[tuple[Any, ...], float]
PartialSpectra = dict[tuple[Any, ...], list[float]]

def alpha_synergistic_entropy(
    joint_distribution: DiscreteDist,
    alpha: int,
    num_samples: int = -1,
    definition: str = "min",
) -> AlphaSynDist:
    """
    Computes the :math:`\\alpha`-synergistic entropy for a joint distribution
    for a given value of :math:`\\alpha`.

    .. math::

        h^{syn}_{\\alpha}(x) = \\min_{a\\subseteq x, |a|=\\alpha} h(x^{a}|x^{-a})


    Parameters
    ----------
    joint_distribution : dict[tuple, float]
        The joint distribution dictionary object.
    alpha : int
        The scale to consider.
    num_samples : int, optional
        The number of samples to trial. The default is -1, in which case, all permutations are trialed.
    definition : str, optional
        How to define the loss of information. Can be "min", "max", or "avg". The default is "min".

    Returns
    -------
    dict[tuple, float]
        The local alpha-synergy for each state.

    References
    ----------
    Varley, T. F. (2024).
    A scalable synergy-first backbone decomposition of
    higher-order structures in complex systems.
    Npj Complexity, 1(1), 1–11.
    https://doi.org/10.1038/s44260-024-00011-1
    
    """
    assert definition in {
        "min",
        "max",
        "avg",
    }, "The optional definitions are 'min', 'max', or 'avg'."

    N: int = len(next(iter(joint_distribution)))

    val: float = 0.0

    if definition == "min":
        val = np.inf
    elif definition == "max":
        val = -np.inf

    alpha_syns: AlphaSynDist = {key: val for key in joint_distribution.keys()}

    sources: list[tuple[Any, ...]] = []
    if num_samples == -1:  # Get all combinations of elements of size alpha.
        sources += list(itertools.combinations(tuple(i for i in range(N)), r=alpha))
    else:  # Randomly sample elements of size alpha.
        sources += list(
            {
                tuple(sorted(np.random.choice(N, size=alpha, replace=False).tolist()))
                for i in range(num_samples)
            }
        )

    num_sources: int = len(sources)

    residuals: list[
        tuple[Any, ...]
    ] = [  # The residual indices for each source in sources.
        tuple(i for i in range(N) if i not in source) for source in sources
    ]

    for i in range(num_sources):
        source: tuple[Any, ...] = sources[i]  # The indices of the source variables.
        complement: tuple[Any, ...] = residuals[
            i
        ]  # The indices of the complenetary variables.

        idxs: tuple[Any, ...] = source + complement

        ptw, _ = conditional_entropy(source, complement, joint_distribution)

        for key in ptw.keys():
            # Undoing the unpacking of state into source and
            # conditional
            flatten: tuple[Any, ...] = sum(key, ())  # ((0,1),(2,)) -> (0,1,2)

            temp: list[int] = [0] * N
            for i in range(N):  # For each element of flatten:
                temp[idxs[i]] = flatten[i]  # Mapping back to the original order.
            state: tuple[Any, ...] = tuple(temp)

            # Accounting for the various ways synergy can be defined.
            if definition == "min":
                if alpha_syns[state] > ptw[key]:
                    alpha_syns[state] = ptw[key]
            elif definition == "max":
                if alpha_syns[state] < ptw[key]:
                    alpha_syns[state] = ptw[key]
            elif definition == "avg":
                alpha_syns[state] += ptw[key] / float(num_sources)

    return alpha_syns


def partial_entropy_spectra(
    joint_distribution: DiscreteDist,
    num_samples: int = -1,
    definition: str = "min",
) -> PartialSpectra:
    """
    Computes the partial synergy for every value of :math:`alpha` (the spectrum) for each local state. 

    Parameters
    ----------
    joint_distribution : dict[tuple,float]
        The joint probability dictionary object.
    num_samples : int, optional
        The number of samples to trial. The default is -1, in which case, all permutations are trialed.
    definition : str, optional
        How to define the loss of information. Can be "min", "max", or "avg". The default is "min".

    Returns
    -------
    dict[tuple,list]
        The alpha-synergistic entropy spectrum for each state.

    """

    N: int = len(list(joint_distribution.keys())[0])
    spectra: dict[tuple[Any, ...], list[float]] = {
        key: [] for key in joint_distribution.keys()
    }

    for alpha in range(1, N + 1):
        alpha_syns: AlphaSynDist = alpha_synergistic_entropy(
            joint_distribution,
            alpha=alpha,
            num_samples=num_samples,
            definition=definition,
        )

        for key in alpha_syns.keys():
            spectra[key].append(alpha_syns[key] - sum(spectra[key]))

    return spectra


def partial_kullback_leibler_spectra(
    posterior: DiscreteDist,
    prior: DiscreteDist,
    num_samples: int = -1,
    definition: str = "min",
) -> PartialSpectra:
    """
    Computes the local Kullback-Leibler spectrum for each state

    Parameters
    ----------
    posterior : dict[tuple,float]
        The distribution that describes the posterior beliefs.
    prior : dict[tuple,float]
        The distribution that describes the prior beliefs.
    num_samples : int, optional
        The number of samples to trial.
        The default is -1, in which case, all permutations are trialed.
    definition : str, optional
        How to define the loss of information.
        Can be "min", "max", or "avg". The default is "min".

    Returns
    -------
    dict[tuple,list]
        The alpha-synergistic DKL spectrum for each state.

    """

    dkl_spectra: PartialSpectra = {key: [] for key in posterior.keys()}
    prior_spectra: PartialSpectra = partial_entropy_spectra(
        prior, num_samples=num_samples, definition=definition
    )
    posterior_spectra: PartialSpectra = partial_entropy_spectra(
        posterior, num_samples=num_samples, definition=definition
    )

    for key in dkl_spectra.keys():
        dkl_spectra[key] = list(
            map(lambda x, y: x - y, prior_spectra[key], posterior_spectra[key])
        )

    return dkl_spectra


def partial_total_correlation_spectra(
    joint_distribution: DiscreteDist,
    num_samples: int = -1,
    definition: str = "min",
) -> PartialSpectra:
    """
    Computes the local total correlation spectrum for each state using the Kullback-Leibler divergence.

    Parameters
    ----------
    joint_distribution : dict[tuple,float]
        The joint probability dictionary object.
    num_samples : int, optional
        The number of samples to trial. The default is -1, in which case, all permutations are trialed.
    definition : str, optional
        How to define the loss of information. Can be "min", "max", or "avg". The default is "min".

    Returns
    -------
    dict[tuple,list]
        The alpha-synergistic total correlation spectrum for each state.


    """

    prior: DiscreteDist = constrained_maximum_entropy_distributions(
        joint_distribution, order=1
    )

    return partial_kullback_leibler_spectra(
        joint_distribution, prior, num_samples=num_samples, definition=definition
    )


def partial_information_spectra(
    inputs: tuple[Any, ...],
    target: tuple[Any, ...],
    joint_distribution: DiscreteDist,
    num_samples: int = -1,
    definition: str = "min",
) -> list[float]:
    """
    Computes the local mutual information spectrum for each state.

    Parameters
    ----------
    inputs : tuple
        The indices of the input variables.
    target : tuple
        The indices of the target variables.
    joint_distribution : dict[tuple,float]
        The joint probability dictionary object.
    num_samples : int, optional
        The number of samples to trial. The default is -1, in which case, all permutations are trialed.
    definition : str, optional
        How to define the loss of information. Can be "min", "max", or "avg". The default is "min".

    Returns
    -------
    list[float]
    """

    N: int = len(inputs)

    # The marginal distribution on the inputs
    # and target.
    marginal_inputs: DiscreteDist = get_marginal_distribution(
        inputs, joint_distribution
    )
    marginal_targets: DiscreteDist = get_marginal_distribution(
        target, joint_distribution
    )

    # The alpha-synergistic entropy decomposition of h(x)
    input_entropy_spectra: PartialSpectra = partial_entropy_spectra(
        marginal_inputs, num_samples=num_samples, definition=definition
    )

    avg: list[float] = [0.0 for _ in range(N)]
    for target_key in marginal_targets.keys():
        p_y: float = marginal_targets[target_key]

        target_conditional_distribution: DiscreteDist = {
            reduce_state(key, inputs): joint_distribution[key] / p_y
            for key in joint_distribution.keys()
            if reduce_state(key, target) == target_key
        }

        conditional_entropy_spectra: PartialSpectra = partial_entropy_spectra(
            target_conditional_distribution,
            num_samples=num_samples,
            definition=definition,
        )

        ptw: list[float] = [0 for _ in range(N)]
        for input_key in conditional_entropy_spectra.keys():
            
            input_spectra: list[float] = input_entropy_spectra[input_key]
            conditional_spectra: list[float] = conditional_entropy_spectra[input_key]
            target_conditional_entropy: float = target_conditional_distribution[input_key]

            for i in range(N):
                ptw[i] += target_conditional_entropy * (
                    input_spectra[i] - conditional_spectra[i]
                )
        
        for i in range(N):
            avg[i] += ptw[i] * p_y
        
    return avg



import scipy.stats as stats
import torch
from .shannon import differential_entropy


def total_correlation(
    idxs: tuple[int],
    data: torch.Tensor,
    context: None | tuple[int] = None,
    data_test: None | torch.Tensor = None,
    flow_kwargs: dict = None,
    train_kwargs: dict = None,
    verbose: bool = False,
) -> tuple[float, float]:
    """
    Computes the total correlation of the data using normalizing flow estimators.  


    Parameters
    ----------
    idxs : tuple[int, ...]
        The tuple of indices the differential entropy is computed for.
    data : torch.Tensor
        The training data, in samples x features format.
    context : None | tuple[int]
        If not None, the indices of the conditioning variables.
    data_test : None | torch.Tensor
        If not None, the testing data in samples x features format.
    flow_kwargs : dict
        Arguments for the utils.initalize_flow function.
    train_kwargs : dict
        Arguments for the utils.train_flow function.
    verbose : bool
        Whether to print the training progress.


    Returns
    -------
    float


    """
    flow_kwargs = flow_kwargs or {}
    train_kwargs = train_kwargs or {}

    if context is None:
        context_arg = None
    else:
        context_arg = context

    lookup: dict[tuple[int,...], float] = {}

    h_idxs, _ = differential_entropy(
        idxs=idxs,
        data=data,
        context=context_arg,
        data_test=data_test,
        verbose=verbose,
        train_kwargs=train_kwargs,
        flow_kwargs=flow_kwargs,
    )
    lookup[idxs] = h_idxs

    for i in idxs:
        h_i, _ = differential_entropy(
            idxs=(i,),
            data=data,
            context=context_arg,
            data_test=data_test,
            verbose=verbose,
            flow_kwargs=flow_kwargs,
            train_kwargs=train_kwargs,
        )

        lookup[(i,)] = h_i

    tc: float = 0.0

    for key in lookup.keys():
        if len(key) == 1:
            tc += lookup[key][0]
        else:
            tc -= lookup[key][0]

    return tc


def higher_order_information(
    idxs: tuple[int],
    data: torch.Tensor,
    context: None | tuple[int] = None,
    data_test: None | torch.Tensor = None,
    flow_kwargs: dict = None,
    train_kwargs: dict = None,
    verbose: bool = False,
) -> dict[str, float]:
    """
    Computes the O-information, S-information, total correlation, and dual total correlation for the data. 
    Computing them all as a set is more efficient than computing each one independently.

    Parameters
    ----------
    idxs : tuple[int, ...]
        The tuple of indices the differential entropy is computed for.
    data : torch.Tensor
        The training data, in samples x features format.
    context : None | tuple[int]
        If not None, the indices of the conditioning variables.
    data_test : None | torch.Tensor
        If not None, the testing data in samples x features format.
    flow_kwargs : dict
        Arguments for the utils.initalize_flow function.
    train_kwargs : dict
        Arguments for the utils.train_flow function.
    verbose : bool
        Whether to print the training progress.
        
        

    Returns
    -------
    dict[str, float]
    
        

    """
    flow_kwargs = flow_kwargs or {}
    train_kwargs = train_kwargs or {}

    if context is None:
        context_arg = None
    else:
        context_arg = context

    lookup_marginals: dict[tuple[int, ...], float] = {(i,): 0.0 for i in idxs}
    lookup_residuals: dict[tuple[int, ...], float] = {tuple(j for j in idxs if j != i): 0.0 for i in idxs}
    
    N: int = len(idxs)
    
    h_idxs, _ = differential_entropy(
        idxs=idxs,
        data=data,
        context=context_arg,
        data_test=data_test,
        verbose=verbose,
        train_kwargs=train_kwargs,
        flow_kwargs=flow_kwargs,
    )
    
    for i in idxs:
        h_i, _ = differential_entropy(
            idxs=(i,),
            data=data,
            context=context_arg,
            data_test=data_test,
            verbose=verbose,
            flow_kwargs=flow_kwargs,
            train_kwargs=train_kwargs,
        )
        lookup_marginals[(i,)] = h_i

        idxs_minus_i = tuple(j for j in idxs if j != i)
        h_minus_i, _ = differential_entropy(
            idxs = idxs_minus_i,
            data = data, 
            context = context_arg,
            data_test = data_test,
            verbose = verbose,
            flow_kwargs = flow_kwargs,
            train_kwargs = train_kwargs
        )
        lookup_residuals[idxs_minus_i] = h_minus_i

    tc_idxs: float =  sum(lookup_marginals[key] for key in lookup_marginals.keys()) - h_idxs
    
    residual_tcs: dict = {}
    for key in lookup_residuals:
        residual_tc = -lookup_residuals[key]
        for i in key:
            residual_tc += lookup_marginals[(i,)]

        residual_tcs[key] = residual_tc

    results = {
        "o_information": -(((N-2)*tc_idxs) - sum(residual_tcs[key] for key in residual_tcs.keys())),
        "s_information": (N*tc_idxs) - sum(residual_tcs[key] for key in residual_tcs.keys()),
        "dual_total_correlation": ((N-1)*tc_idxs) - sum(residual_tcs[key] for key in residual_tcs.keys()),
        "total_correlation": tc_idxs 
    }

    return results 



# import pickle
import numpy as np
import itertools


def make_powerset(iterable):
    """
    A utility function for quickly making powersets,

    powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)

    """
    xs: list = list(iterable)
    # note we return an iterator rather than a list
    return itertools.chain.from_iterable(itertools.combinations(xs, n) for n in range(len(xs) + 1))


def clean_distribution(joint_distribution: dict[tuple, float]) -> dict:
    """
    A utility function to remove states with 0 probability

    Parameters
    ----------
    joint_distribution: dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    dict
        The joint probability distribution with zero-probability elements removed.

    """
    return {
        key: joint_distribution[key]
        for key in joint_distribution.keys()
        if joint_distribution[key] > 0.0
    }


def reduce_state(state: tuple, source: tuple) -> tuple:
    """
    A utility function for reducing tuples
    to just the elements in the source.

    Parameters
    ----------
    state : tuple
        The particular state of each variable.
    source : tuple
        The indices of the variable to remove.

    Returns
    -------
    tuple
        The reduced state consisting only of those
        elements indexed in the source variable.

    """

    return tuple(state[i] for i in source)


def construct_joint_distribution(data: np.ndarray) -> dict[tuple, float]:
    """
    Given a channels x time, discrete Numpy array, computes
    the probability distribution that describes the data.


    Parameters
    ----------
    data : np.ndarray
        The data: assumed to be in elements x time format.

    Returns
    -------
    dict
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    """

    assert data.dtype != "float", "The array must be discrete-valued variables."

    N0: int
    N1: int
    N0, N1 = data.shape

    unq: np.ndarray
    counts: np.ndarray
    unq, counts = np.unique(data, return_counts=True, axis=-1)

    return {tuple(unq[:, i]): counts[i] / counts.sum() for i in range(unq.shape[1])}


def get_marginal_distribution(
    idxs: tuple, joint_distribution: dict[tuple, float]
) -> dict:
    """
    Returns the marginal distribution of the variables
    indexed by the idxs tuple. The opposite of the
    marginalize_out() function.

    Parameters
    ----------
    idxs : tuple
        The indices of the variable to retain.
    joint_distribution: dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    dict
        The marginal joint probability distribution object.

    """

    states: list = list(joint_distribution.keys())

    reduced_states: set = {reduce_state(state, idxs) for state in states}
    reduced_distribution: dict = {r_state: 0.0 for r_state in reduced_states}

    for r_state in reduced_states:
        reduced_distribution[r_state] = sum(
            joint_distribution[state]
            for state in joint_distribution.keys()
            if reduce_state(state, idxs) == r_state
        )

    return reduced_distribution


def marginalize_out(idxs: tuple, joint_distribution: dict[tuple, float]) -> dict:
    """
    Returns a distribution with the variables indexed by
    idxs marginalized out.

    Parameters
    ----------
    idxs : tuple
        The indices of the variables to be marginalized out.
    joint_distribution: dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    dict
        A joint probability distribution dictionary.

    """

    N: int = len(list(joint_distribution.keys())[0])

    residuals: tuple = tuple(i for i in range(N) if i not in idxs)

    return get_marginal_distribution(residuals, joint_distribution)


def get_all_marginal_distributions(
    joint_distribution: dict[tuple, float],
) -> dict[tuple, dict]:
    """
    Computes the set of all marginal probability distributions.
    If the original distribution has variables:

        :math:`P(X_1, X_2, X_3)`

    Returns a dictionary of dictionaries for each:
        :math:`P(X_1,), P(X_2,), P(X_3,), P(X_1, X_2), P(X_1, X_3), P(X_2, X_3), P(X_1,X_2,X_3)`


    Parameters
    ----------
    joint_distribution: dict[tuple, float][tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    dict[tuple, dict]
        A dictionary of dictionaries: each key is a set of marginals,
        each value is the associated marginal distribution .

    """

    N: int = len(list(joint_distribution.keys())[0])

    sources: list = list(make_powerset(range(N)))
    sources.remove(())

    marginal_dict: dict = {
        source: get_marginal_distribution(source, joint_distribution)
        for source in sources
    }

    return marginal_dict


def product_distribution(
    A: dict[tuple[int, ...], float], B: dict[tuple[int, ...], float]
) -> dict[tuple[int, ...], float]:
    """
    Compute the product of two independent distributions.

    Parameters
    ----------
    A : dict[tuple[int, ...], float]
        The first distribution
    B : dict[tuple[int, ...], float]
        The second distribution

    Returns
    -------
    dict[tuple[int, ...], float]
        Joint distribution over concatenated states
    """
    result = {}
    for state_a, prob_a in A.items():
        for state_b, prob_b in B.items():
            joint_state = state_a + state_b
            result[joint_state] = prob_a * prob_b
    return result


def generate_closed_distribution(N: int, seed: int = None) -> dict[tuple[int, ...], float]:
    """
    Generate a random closed discrete probability distribution on N binary elements.
    
    A distribution is closed iff H(X_i | X^{-i}) = 0 for all i, meaning every
    variable is fully determined by the others. This requires the support to
    have minimum Hamming distance >= 2.
    
    Parameters
    ----------
    N : int
        Number of binary elements.
    seed : int, optional
        Random seed for reproducibility.
        
    Returns
    -------
    dict[tuple[int, ...], float]
        Probability distribution as {state: probability} mapping.
    """
    rng = np.random.default_rng(seed)
    
    # Generate all 2^N possible states
    all_states = list(itertools.product([0, 1], repeat=N))
    
    # Build valid support: greedily add states with Hamming distance >= 2 from all others
    support = []
    candidates = list(all_states)
    rng.shuffle(candidates)
    
    for state in candidates:
        # Check if this state has Hamming distance >= 2 from all states in support
        valid = True
        for s in support:
            hamming_dist = sum(a != b for a, b in zip(state, s))
            if hamming_dist < 2:
                valid = False
                break
        if valid:
            support.append(state)
    
    # Assign random probabilities to support states
    weights = rng.random(len(support))
    weights /= weights.sum()
    
    return {state: float(p) for state, p in zip(support, weights)}



import numpy as np
import itertools as it
from .utils import get_marginal_distribution, reduce_state


def constrained_maximum_entropy_distributions(
    joint_distribution: dict[tuple[int, ...], float],
    marginal_constraints: list[tuple] = [(None,)],
    order: int = -1,
    max_iters: int = 10_000,
    tol: float = 1e-6,
) -> dict[tuple, float]:
    """
    Uses the iterated proportional fitting (IPF) algorithm to find
    the maximum-entropy distribution consistant with given constraints.

    If the marginal constrains are given, uses those.
    If order is given, finds the maximum entropy distribution consistant
    with all marginals of the given order.

    Parameters
    ----------
    joint_distribution : dict
        The joint distribution from which to compute the marginal constraints.
    marginal_constraints : list, optional
        A list of tuples: each tuple corresponds to a set of marginals to constrain. The default is (None,).
    order : int, optional
        The order of marginals to fix. The default is -1.
    max_iters : int, optional
        The maximum number of iterations the algorithm can run for. The default is 10_000.
    tol : float, optional
        The value below which further refinements of the maximum entropy distribution are stopped.
        The default is 1e-6.

    Returns
    -------
    dict[tuple, float]
        The optimized distribution consistent with the given marginal constraints.

    """

    assert (marginal_constraints == [(None,)]) ^ (
        order == -1
    ), "Give just one: fixed constraints or marginal order."

    N = len(list(joint_distribution.keys())[0])

    if order != -1:
        marginal_constraints = list(it.combinations(range(N), r=order))

    marginals = {
        m: get_marginal_distribution(m, joint_distribution)
        for m in marginal_constraints
    }

    first_order_marginals = [
        get_marginal_distribution((i,), joint_distribution) for i in range(N)
    ]
    first_order_states = [list(m.keys()) for m in first_order_marginals]
    # Evil tuple unpacking. This works but I hate it.
    first_order_states = [
        tuple(item for tuple in first_order_states[x] for item in tuple)
        for x in range(len(first_order_states))
    ]

    maxent_states = list(it.product(*first_order_states))

    if (
        order == 1
    ):  # If the order is 1, return the product of the first-order marginals.
        maxent = {}

        for state in maxent_states:
            maxent[state] = np.prod(
                [first_order_marginals[i][(state[i],)] for i in range(N)]
            )

        # return maxent
    else:  # Otherwise do the IPF algorithm.
        maxent = {state: 1 / len(maxent_states) for state in maxent_states}

        for _ in range(max_iters):

            prev = maxent.copy()
            for m in marginal_constraints:

                target_marg = marginals[m]
                maxent_marg = get_marginal_distribution(m, maxent)

                scaling_factors = {
                    key: (
                        target_marg[key] / maxent_marg[key]
                        if maxent_marg[key] > 0
                        else 0
                    )
                    for key in target_marg.keys()
                }

                maxent = {
                    state: (
                        maxent[state] * scaling_factors[reduce_state(state, m)]
                        if reduce_state(state, m) in scaling_factors
                        else maxent[state]
                    )
                    for state in maxent.keys()
                }
                maxent = {
                    state: maxent[state] / sum(maxent.values())
                    for state in maxent.keys()
                }

            if sum([abs(prev[key] - maxent[key]) for key in maxent.keys()]) < tol:
                break

    return maxent



from .shannon import (
    shannon_entropy,
    conditional_entropy,
    mutual_information,
    conditional_mutual_information,
    kullback_leibler_divergence,
)

from .multivariate_mi import (
    total_correlation,
    dual_total_correlation,
    s_information,
    o_information,
    tse_complexity,
    description_complexity,
    co_information,
    connected_information,
)

from .decompositions import (
    partial_entropy_decomposition,
    partial_information_decomposition,
    generalized_information_decomposition,
    representational_complexity,
)

from .temporal import (
    lempel_ziv_complexity,
    lempel_ziv_mutual_information,
    lempel_ziv_total_correlation,
    cross_lempel_ziv_complexity,
)

__all__ = [
    "shannon_entropy",
    "conditional_entropy",
    "mutual_information",
    "conditional_mutual_information",
    "kullback_leibler_divergence",
    "total_correlation",
    "dual_total_correlation",
    "s_information",
    "o_information",
    "tse_complexity",
    "description_complexity",
    "co_information",
    "connected_information",
    "partial_entropy_decomposition",
    "partial_information_decomposition",
    "generalized_information_decomposition",
    "representational_complexity",
    "lempel_ziv_complexity",
    "lempel_ziv_mutual_information",
    "lempel_ziv_total_correlation",
    "cross_lempel_ziv_complexity"
]



import numpy as np
from .utils import get_marginal_distribution

from typing import Any

DiscreteDist = dict[tuple[Any, ...], float]

def shannon_entropy(joint_distribution: DiscreteDist) -> tuple[dict, float]:
    """
    Computes the Shannon entropy of the distribution :math:`P(x)`.

    .. math::
        H(X) = -\\sum_{x\\in\\mathcal{X}} P(x) \\log P(x)

    To compute the entropy of a subset of the variables in the joint
    distribution, use the :func:`syntropy.discrete.utils.get_marginals` function from the utils
    library.

    Parameters
    ----------
    joint_distribution : dict
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    ptw : dict
        The pointwise entropy for each state in the joint distribution.
    avg : float
        The average entropy

    """

    ptw: dict = {
        state: -np.log2(joint_distribution[state])
        for state in joint_distribution.keys()
        if joint_distribution[state] > 0.0
    }
    avg: float = sum(joint_distribution[state] * ptw[state] for state in ptw.keys())

    return ptw, avg


def conditional_entropy(
    idxs_x: tuple[int, ...], idxs_y: tuple[int, ...], joint_distribution: DiscreteDist
) -> tuple[dict, float]:
    """
    Computes the conditional entropy of X given Y.

    .. math::
        H(X|Y) = H(X,Y) - H(Y)

    Parameters
    ----------
    idxs_x : tuple
        The indices of the variables to compute the entropy on.
    idxs_y : tuple
        The indicies of the variables to contintue on.
    joint_distribution : dict
        DESCRIPTION.joint_distribution : dict
            The joint probability distribution.
            Keys are tuples corresponding to the state of each element.
            The valules are the probabilities.

    Returns
    -------
    ptw : dict
        The pointwise entropy for each state in the joint distribution.
    avg : float
        The average entropy

    """

    Nx: int = len(idxs_x)

    idxs_xy: tuple[int, ...] = idxs_x + idxs_y

    marginal_xy: DiscreteDist = get_marginal_distribution(idxs_xy, joint_distribution)
    marginal_y: DiscreteDist = get_marginal_distribution(idxs_y, joint_distribution)

    ptw: dict = {}
    avg: float = 0

    for state in marginal_xy.keys():
        if marginal_xy[state] > 0.0:
            p_y: float = marginal_y[state[Nx:]]
            h: float = -np.log2(marginal_xy[state] / p_y)
            ptw[((state[:Nx]), (state[Nx:]))] = h

            avg += marginal_xy[state] * h

    return ptw, avg


def mutual_information(
    idxs_x: tuple[int, ...],
    idxs_y: tuple[int, ...],
    joint_distribution: DiscreteDist,
) -> tuple[dict, float]:
    """
    Computes the mutual information between X and Y.

    .. math:: 
        I(X;Y) &= H(X) + H(Y) - H(X,Y) \\\\
               &= H(X) - H(X|Y) \\\\
               &= H(Y) - H(Y|X) \\\\
               &= H(X,Y) - H(X|Y) - H(Y|X)

    Parameters
    ----------
    idxs_x : tuple
        The indices of the X variable(s).
    idxs_y : tuple
        The indices of the Y variable(s).
    joint_distribution : dict
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.


    Returns
    -------
    ptw : dict
        The pointwise mutual information for each state in the joint distribution.
    avg : float
        The average mutual information

    """

    Nx: int = len(idxs_x)

    idxs_xy: tuple[int, ...] = idxs_x + idxs_y

    marginal_xy: DiscreteDist = get_marginal_distribution(idxs_xy, joint_distribution)
    marginal_y: DiscreteDist = get_marginal_distribution(idxs_y, joint_distribution)
    marginal_x: DiscreteDist = get_marginal_distribution(idxs_x, joint_distribution)

    ptw: dict = {}
    avg: float = 0.0

    for state in marginal_xy.keys():
        if marginal_xy[state] > 0.0:
            p_x: float = marginal_x[state[:Nx]]
            p_y: float = marginal_y[state[Nx:]]

            mi = np.log2(marginal_xy[state] / (p_x * p_y))

            ptw[((state[:Nx]), (state[Nx:]))] = mi

            avg += mi * marginal_xy[state]

    return ptw, avg


def conditional_mutual_information(
    idxs_x: tuple, idxs_y: tuple, idxs_z: tuple, joint_distribution: DiscreteDist
) -> tuple[dict, float]:
    """
    Computes the mutual information between X and Y condioned on Z.

    .. math:: 
        I(X,Y|Z) &= H(X|Z) + H(Y|Z) - H(X,Y|Z) \\\\
                 &= I(X;Y,Z) - I(X;Z)

    Parameters
    ----------
    idxs_x : tuple
        The indices of the X variable(s).
    idxs_y : tuple
        The indices of the Y variable(s).
    idxs_z : tuple
        The indices of the variables to condition on.
    joint_distribution : dict
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    ptw : dict
        The pointwise mutual information for each state in the joint distribution.
    avg : float
        The average mutual information

    """

    Nx: int = len(idxs_x)

    joint: tuple[int, ...] = idxs_x + idxs_y

    ptw_xz, avg_xz = conditional_entropy(idxs_x, idxs_z, joint_distribution)
    ptw_yz, avg_yz = conditional_entropy(idxs_y, idxs_z, joint_distribution)
    ptw_xyz, avg_xyz = conditional_entropy(joint, idxs_z, joint_distribution)

    avg: float = avg_xz + avg_yz - avg_xyz
    ptw: dict = {}

    for state in ptw_xyz.keys():
        sx = state[0][:Nx]
        sy = state[0][Nx:]

        sxz = (sx, state[1])
        syz = (sy, state[1])

        ptw[(sx, sy, state[1])] = ptw_xz[sxz] + ptw_yz[syz] - ptw_xyz[state]

    return ptw, avg


def kullback_leibler_divergence(
    posterior_distribution: DiscreteDist, prior_distribution: DiscreteDist
) -> tuple[dict, float]:
    """
    Computes the Kullback-Leibler divergence from a prior distribution P(X) and
    and posterior distribution Q(X).

    .. math:: 
        D_{KL}(P||Q) = \\sum_{x} P(x) \\log \\frac{P(x)}{Q(x)}

    Parameters
    ----------
    posterior_distribution : dict
        The joint distribution of the posterior distribution P(X).
    prior_distribution : dict
        The joint distribution of the prior distribution Q(X)

    Returns
    -------
    ptw : dict
        The pointwise Kullback-Leibler divergence for each state in the joint distribution.
    avg : float
        The average Kullback-Leibler divergence.

    """

    assert set(prior_distribution.keys()).issuperset(
        set(posterior_distribution.keys())
    ), "The support set of the prior must be a superset of the posterior"

    avg: float = 0
    ptw: dict = {state: 0 for state in posterior_distribution.keys()}
    for state in posterior_distribution.keys():
        log_ratio: float = np.log2(
            posterior_distribution[state] / prior_distribution[state]
        )

        avg += posterior_distribution[state] * log_ratio
        ptw[state] = log_ratio

    return ptw, avg



"""
Syntropy: A package for multivariate information theory on discrete and continuous random variables.

This package provides tools for computing information-theoretic quantities
including entropy, mutual information, and various multivariate decompositions
for discrete, Gaussian, and general continuous distributions.
"""

__version__ = "0.0.1"
__author__ = "Thomas F. Varley"
__email__ = "tfvarley@uvm.edu"

from . import discrete
from . import gaussian
from . import knn
from . import neural
from . import mixed
from . import lattices

__all__ = [
    "discrete",
    "gaussian",
    "knn",
    "neural",
    "mixed",
    "lattices",
    "__version__",
    "__author__",
    "__email__",
]



import numpy as np
from numpy.typing import NDArray
from scipy.spatial import cKDTree


def check_idxs(idxs: tuple[int, ...], N: int) -> tuple[int, ...]:
    """
    Checks whether the idxs input is -1, in which case, defaults to all processes.

    Parameters
    ----------
    idxs : tuple[int, ...]
        Indices of variables to use (-1 means all)
    N : int
        The total number of processes.

    Returns
    -------
    tuple[int, ...]
        Either idxs, or a tuple of all processes.

    """
    if idxs[0] == -1:
        idxs_ = tuple(i for i in range(N))
    else:
        idxs_ = idxs

    return idxs_


def build_tree_and_get_distances(
    data: NDArray[np.floating], k: int
) -> tuple[cKDTree, NDArray[np.floating], NDArray[np.integer]]:
    """
    Builds the KNN tree and returns the indices and distances between each point and it's k-nearest neighbors.

    Parameters
    ----------
    data : NDArray[np.floating]
        Data array of shape (n_variables, n_samples)
    k : int
        Number of nearest neighbors

    Returns
    -------
    cKDTree
        The KNN tree constructed from data.
    NDArray[np.floating]
        The indices of each of the k-nearest neighbors.
    NDArray[np.integer]
        The distances to each k-nearest neighbors (using the max norm).

    """
    tree = cKDTree(data.T)

    distances: NDArray[np.floating]
    indices: NDArray[np.integer]
    distances, indices = tree.query(data.T, k=k + 1, p=np.inf)

    return tree, distances, indices


def get_counts_from_tree(
    tree: cKDTree,
    data: NDArray[np.floating],
    eps: NDArray[np.floating],
    strict: bool = True,
) -> NDArray[np.integer]:

    if strict is True:
        eps_ = np.nextafter(eps, -np.inf)
    else:
        eps_ = eps
    counts: NDArray[np.integer] = np.array(
        [
            tree.query_ball_point(x_i, eps_i, return_length=True, p=np.inf) - 1
            for x_i, eps_i in zip(data.T, eps_)
        ]
    )

    return counts



import pathlib
import pickle 

lattice_path = pathlib.Path(__file__).parent

with open(lattice_path / 'pi_lattice_2.pickle', 'rb') as f:
    LATTICE_2 = pickle.load(f)
with open(lattice_path / 'pi_lattice_3.pickle', 'rb') as f:
    LATTICE_3 = pickle.load(f)
with open(lattice_path / 'pi_lattice_4.pickle', 'rb') as f:
    LATTICE_4 = pickle.load(f)



# test_imports.py
print("Testing discrete...")
import syntropy.discrete
print("✓ discrete OK")

print("Testing gaussian...")
import syntropy.gaussian  
print("✓ gaussian OK")

print("Testing knn...")
import syntropy.knn
print("✓ knn OK")

print("Testing neural...")
import syntropy.neural
print("✓ neural OK")

print("\n✓ All imports successful - no circular dependencies!")



import torch
import nflows
from .utils import initialize_flow, train_flow, evaluate_flow
from typing import Any


def differential_entropy(
    idxs: tuple[int, ...],
    data: torch.Tensor,
    data_test: None | torch.Tensor = None,
    flow_kwargs: None | dict[str, Any] = None,
    train_kwargs: None | dict[str, Any] = None,
    verbose: bool = False,
) -> tuple[torch.Tensor, float, nflows.flows.base.Flow]:
    """
    Computes the differential entropy of the data.

    Parameters
    ----------
    idxs : tuple[int, ...]
        The tuple of indices the differential entropy is computed for.
    data : torch.Tensor
        The training data, in samples x features format.
    data_test : None | torch.Tensor
        If not None, the testing data in samples x features format.
    flow_kwargs : dict
        Arguments for the utils.initalize_flow function.
    train_kwargs : dict
        Arguments for the utils.train_flow function.
    verbose : bool
        Whether to print the training progress.


    Returns
    -------
    float

    nflows.flows.base.Flow


    """
    flow_kwargs = flow_kwargs or {}
    train_kwargs = train_kwargs or {}

    # initialize the flow
    dim: int = len(idxs)
    flow = initialize_flow(dim=dim, **flow_kwargs)

    # train the flow
    flow = train_flow(
        flow=flow,
        data=data[:, idxs],
        verbose=verbose,
        **train_kwargs,
    )
    # evaluation
    eval_data = data if data_test is None else data_test
    ptw, avg = evaluate_flow(flow, eval_data[:, idxs])

    return ptw, avg, flow


def conditional_entropy(
    idxs_x: tuple[int, ...],
    idxs_y: tuple[int, ...],
    data: torch.Tensor,
    data_test: None | torch.Tensor = None,
    flow_kwargs: None | dict[str, Any] = None,
    train_kwargs: None | dict[str, Any] = None,
    verbose: bool = False,
) -> tuple[torch.Tensor, float, nflows.flows.base.Flow]:
    """
    Computes the differential entropy of the data.

    Parameters
    ----------
    idxs_x : tuple[int, ...]
        The tuple of indices the differential entropy is computed for.
    idxs_y : tuple[int, ...]
        The tuple of indices to condition on.
    data : torch.Tensor
        The training data, in samples x features format.
    data_test : None | torch.Tensor
        If not None, the testing data in samples x features format.
    flow_kwargs : dict
        Arguments for the utils.initalize_flow function.
    train_kwargs : dict
        Arguments for the utils.train_flow function.
    verbose : bool
        Whether to print the training progress.


    Returns
    -------
    torch.Tensor

    float

    nflows.flows.base.Flow


    """
    flow_kwargs = flow_kwargs or {}
    train_kwargs = train_kwargs or {}

    flow_kwargs["dim_context"] = len(idxs_y)

    # initialize the flow
    dim: int = len(idxs_x)
    flow = initialize_flow(dim=dim, **flow_kwargs)

    # train the flow
    flow = train_flow(
        flow=flow,
        data=data[:, idxs_x],
        context=data[:, idxs_y],
        verbose=verbose,
        **train_kwargs,
    )
    # evaluation
    eval_data = data if data_test is None else data_test
    
    ptw: torch.Tensor
    avg: float

    ptw, avg = evaluate_flow(flow, eval_data[:, idxs_x], context=eval_data[:, idxs_y])

    return ptw, avg, flow


def mutual_information(
    idxs_x: tuple[int, ...],
    idxs_y: tuple[int, ...],
    data: torch.Tensor,
    data_test: None | torch.Tensor = None,
    flow_kwargs: None | dict[str, Any] = None,
    train_kwargs: None | dict[str, Any] = None,
    verbose: bool = False,
) -> float:
    """
    Computes the mutual information between the elements given by idxs_x and idxs_y.

    Parameters
    ----------
    idxs_x : tuple[int, ...]
        The tuple of indices the x variable.
    idxs_y : tuple[int, ...]
        The tuple of indices for the y variable.
    data : torch.Tensor
        The training data, in samples x features format.
    context : None | tuple[int]
        If not None, the indices of the conditioning variables.
    data_test : None | torch.Tensor
        If not None, the testing data in samples x features format.
    flow_kwargs : dict
        Arguments for the utils.initalize_flow function.
    train_kwargs : dict
        Arguments for the utils.train_flow function.
    verbose : bool
        Whether to print the training progress.


    Returns
    -------
    float


    """
    flow_kwargs = flow_kwargs or {}
    train_kwargs = train_kwargs or {}

    ptw_x, avg_x, _ = differential_entropy(
        idxs=idxs_x,
        data=data,
        data_test=data_test,
        verbose=verbose,
        flow_kwargs=flow_kwargs,
        train_kwargs=train_kwargs,
    )
    ptw_cond, avg_cond, _ = conditional_entropy(
        idxs_x=idxs_x,
        idxs_y=idxs_y,
        data=data,
        data_test=data_test,
        flow_kwargs=flow_kwargs,
        train_kwargs=train_kwargs,
        verbose=verbose,
    )

    ptw: torch.Tensor = ptw_x - ptw_cond
    mi: float = avg_x - avg_cond

    return ptw, mi


def conditional_mutual_information(
    idxs_x: tuple[int, ...],
    idxs_y: tuple[int, ...],
    idxs_z: tuple[int, ...],
    data: torch.Tensor,
    data_test: None | torch.Tensor = None,
    flow_kwargs: None | dict[str, Any] = None,
    train_kwargs: None | dict[str, Any] = None,
    verbose: bool = False,
) -> float:
    """
    Computes the mutual information between the elements given by idxs_x and idxs_y.

    Parameters
    ----------
    idxs_x : tuple[int, ...]
        The tuple of indices the x variable.
    idxs_y : tuple[int, ...]
        The tuple of indices for the y variable.
    data : torch.Tensor
        The training data, in samples x features format.
    context : None | tuple[int]
        If not None, the indices of the conditioning variables.
    data_test : None | torch.Tensor
        If not None, the testing data in samples x features format.
    flow_kwargs : dict
        Arguments for the utils.initalize_flow function.
    train_kwargs : dict
        Arguments for the utils.train_flow function.
    verbose : bool
        Whether to print the training progress.


    Returns
    -------
    float


    """
    flow_kwargs = flow_kwargs or {}
    train_kwargs = train_kwargs or {}

    ptw_x_given_z, avg_x_given_z, _ = conditional_entropy(
        idxs_x=idxs_x,
        idxs_y=idxs_z,
        data=data,
        data_test=data_test,
        verbose=verbose,
        flow_kwargs=flow_kwargs,
        train_kwargs=train_kwargs,
    )
    ptw_x_given_yz, avg_x_given_yz, _ = conditional_entropy(
        idxs_x=idxs_x,
        idxs_y=idxs_y + idxs_z,
        data=data,
        data_test=data_test,
        flow_kwargs=flow_kwargs,
        train_kwargs=train_kwargs,
        verbose=verbose,
    )

    ptw: torch.Tensor = ptw_x_given_z - ptw_x_given_yz
    avg: float = avg_x_given_z - avg_x_given_yz

    return ptw, avg



import pytest
import pathlib
import numpy as np
import pandas as pd

from syntropy.knn import mutual_information, total_correlation, dual_total_correlation

data_path = pathlib.Path(__file__).parent
data = pd.read_csv(data_path / "../examples/bold.csv", header=None).values

pytest_abs = 1e-6
def test_total_correlation():
    tc1 = total_correlation(data=data, k=5, idxs=(0, 1), algorithm=1)
    mi1 = mutual_information(idxs_x=(0,), idxs_y=(1,), data=data, k=5, algorithm=1)

    assert tc1[1] == pytest.approx(mi1[1], abs=pytest_abs)
    tc2 = total_correlation(data=data, k=5, idxs=(0, 1), algorithm=2)
    mi2 = mutual_information(idxs_x=(0,), idxs_y=(1,), data=data, k=5, algorithm=2)

    assert tc1[1] == pytest.approx(mi1[1], abs=pytest_abs)

def test_dual_total_correlation():
    dtc1 = dual_total_correlation(data=data, k=5, idxs=(0, 1))
    mi1 = mutual_information(idxs_x=(0,), idxs_y=(1,), data=data, k=5, algorithm=1)

    assert dtc1[1] == pytest.approx(mi1[1], abs=pytest_abs)



import pytest
import pathlib

import numpy as np
import pandas as pd
import scipy.stats as stats

import syntropy.gaussian.shannon as shannon
import syntropy.gaussian.multivariate_mi as mi
from syntropy.gaussian.decompositions import (
    partial_information_decomposition as pid,
    partial_entropy_decomposition as ped,
    generalized_information_decomposition as gid,
)
from syntropy.gaussian.temporal import (
    differential_entropy_rate,
    mutual_information_rate,
    total_correlation_rate,
    o_information_rate
)

data_path = pathlib.Path(__file__).parent
data = pd.read_csv(data_path / "../examples/bold.csv", header=None).values

cov = np.cov(data, ddof=0.0)

# Due to natural instability in Scipy's matrix algebra, we need a slightly
# more relaxed tolerance for our unit tests. 1 part in 1,000,000 is probably ok.
pytest_abs = 1e-6


def test_differential_entropy():

    h = shannon.local_differential_entropy(data[1, :]).mean()
    assert shannon.differential_entropy(cov[1, 1], (1,)) == pytest.approx(
        h, abs=pytest_abs
    )

    h1 = shannon.local_differential_entropy(
        data[(1, 2), :], cov[(1, 2), :][:, (1, 2)]
    ).mean()
    h2 = shannon.local_differential_entropy(
        data[(1, 2), :],
    ).mean()
    assert h1 == pytest.approx(h2, abs=pytest_abs)

    assert h1 == pytest.approx(
        shannon.differential_entropy(cov, (1, 2)), abs=pytest_abs
    )

    c = shannon.conditional_entropy((2,), (1,), cov)

    assert h1 - h == pytest.approx(c, abs=pytest_abs)


def test_mutual_information():

    i = shannon.mutual_information((0,), (1,), cov)
    r, _ = stats.pearsonr(data[0, :], data[1, :])

    assert i == pytest.approx(-np.log(1 - (r**2)) / 2, abs=pytest_abs)

    lmi = shannon.local_mutual_information((0,), (1,), data)
    assert i == pytest.approx(lmi.mean(), abs=pytest_abs)

    tc = mi.total_correlation(cov[(0, 1), :][:, (0, 1)])
    assert i == pytest.approx(tc, abs=pytest_abs)


def test_higher_order_mi():

    idxs = (0, 1, 2, 3)

    tc = mi.total_correlation(cov, idxs)
    dtc = mi.dual_total_correlation(cov, idxs)
    o = mi.o_information(cov, idxs)
    s = mi.s_information(cov, idxs)

    assert tc - dtc == pytest.approx(o, abs=pytest_abs)
    assert tc + dtc == pytest.approx(s, abs=pytest_abs)

    ltc = mi.local_total_correlation(data, cov, idxs)
    ldtc = mi.local_dual_total_correlation(data, cov, idxs)
    lo = mi.local_o_information(data, cov, idxs)
    ls = mi.local_s_information(data, cov, idxs)

    assert ltc.mean() == pytest.approx(tc, abs=pytest_abs)
    assert ldtc.mean() == pytest.approx(dtc, abs=pytest_abs)
    assert lo.mean() == pytest.approx(o, abs=pytest_abs)
    assert ls.mean() == pytest.approx(s, abs=pytest_abs)

    c = mi.description_complexity(cov, idxs)
    assert c == pytest.approx(dtc / len(idxs), abs=pytest_abs)

    triad = np.array(
        [
            [1.0, 0.14621677, 0.59480877],
            [0.14621677, 1.0, 0.27645032],
            [0.59480877, 0.27645032, 0.99999999],
        ]
    )

    mi_12_3 = shannon.mutual_information((0, 1), (2,), triad)
    mi_1_3 = shannon.mutual_information((0,), (2,), triad)
    mi_2_3 = shannon.mutual_information((1,), (2,), triad)

    # Equivavlance between O-information and WMS MI for 3 variables.
    wms = mi_12_3 - (mi_1_3 + mi_2_3)

    assert -wms == pytest.approx(mi.o_information(triad), abs=pytest_abs)

    # Testing the definition in terms of entropy
    whole = shannon.differential_entropy(triad)
    sum_diffs = 0.0
    for i in range(3):
        residuals = tuple(j for j in range(3) if j != i)
        sum_diffs += shannon.differential_entropy(
            triad, idxs=(i,)
        ) - shannon.differential_entropy(triad, idxs=residuals)

    assert whole + sum_diffs == pytest.approx(mi.o_information(triad), abs=pytest_abs)


def test_pid():

    idxs = (0, 1)
    target = (3, 4)
    ptw, avg = pid(idxs, target, data, cov)
    mi = shannon.mutual_information(idxs, target, cov)

    assert sum(avg.values()) == pytest.approx(mi, abs=pytest_abs)

    mi = shannon.mutual_information((0,), target, cov)
    assert avg[((0,),)] + avg[((0,), (1,))] == pytest.approx(mi, abs=pytest_abs)


def test_ped():

    idxs = (0, 1, 2, 3)

    ptw, avg = ped(idxs, data, cov)
    h = shannon.differential_entropy(cov, idxs)

    assert sum(avg.values()) == pytest.approx(h, abs=pytest_abs)


def test_gid():

    idxs = (0, 1, 2, 3)
    cov_prior = np.eye(len(idxs))
    cov_posterior = cov[np.ix_(idxs, idxs)]

    ptw, avg = gid(idxs, data, cov_posterior, cov_prior)

    dkl = shannon.kullback_leibler_divergence(cov_posterior, cov_prior)
    assert sum(avg.values()) == pytest.approx(dkl, abs=pytest_abs)

    ldkl = shannon.local_kullback_leibler_divergence(
        cov_posterior, cov_prior, data[idxs, :]
    )

    assert ldkl.mean() == pytest.approx(sum(avg.values()), abs=pytest_abs)


def test_gaussian_rate():

    T = 5_000_000

    noise = np.random.randn(2, T)

    _, mir = mutual_information_rate((0,), (1,), noise, nperseg=2**13)

    _, h1 = differential_entropy_rate((0,), noise, nperseg=2**13)
    _, h2 = differential_entropy_rate((1,), noise, nperseg=2**13)
    _, h12 = differential_entropy_rate((0, 1), noise, nperseg=2**13)

    assert mir == pytest.approx(h1 + h2 - h12, abs=pytest_abs)

    _, tcr = total_correlation_rate((0, 1), noise, nperseg=2**13)

    assert mir == pytest.approx(tcr, abs=pytest_abs)

    # for white noise with 0 mean and unit variance,
    # the analytic entropy rate should be (1/2)*log(2*pi*e)
    analytic = (1 / 2) * np.log(2.0 * np.pi * np.e)

    assert analytic == pytest.approx(h1, abs=1e-3)

    # %% A VAR(1) model with fixed noise and coefficient.

    X = np.zeros((1, T))
    eps = 0.01
    noise = np.random.randn(T) * eps
    a = 0.2

    for t in range(1, T):
        X[0, t] = X[0, t - 1] * a + noise[t - 1]

    _, hX = differential_entropy_rate((0,), X, nperseg=2**14)

    analytic = 0.5 * (np.log(2 * np.pi * np.e * (eps**2))) - (0.5 * np.log(1 - (a**2)))

    assert np.isclose(hX, analytic, rtol=1e-2)

def test_oinfo_rate():

    T = 5_000_000

    noise = np.random.randn(3, T)
    
    _, mi_joint = mutual_information_rate((0,1), (2,), noise, nperseg=2**13)
    _, mi_1 = mutual_information_rate((0,),(2,), noise, nperseg=2**13)
    _, mi_2 = mutual_information_rate((1,),(2,), noise, nperseg=2**13)

    _, oir = o_information_rate((0,1,2), noise, nperseg=2**13)

    assert oir == pytest.approx(mi_1 + mi_2 - mi_joint, abs=pytest_abs)




import numpy as np
from typing import Any

def lempel_ziv_complexity(
    X: list[Any], return_dictionary: bool = False
) -> float | tuple[float, set]:
    """
    Uses the classic Lempel-Ziv compression algorithm to estimate the entropy rate of a one-dimensional array with :math:`N` samples. 
    If each element in X is a multi-dimensional tuple, then the result is equivalent to the joint entropy rate.

    The extension to multivariate Lempel-Ziv is straightforward and involves representing the joint state of each element at time t as a tuple (X(t), Y(t)) and treating the two sources as a single joint source.

    Here, the dictionary length :math:`|D|` is normalized:

    .. math::
        \\textnormal{Complexity}(X) = \\frac{|D|\\log|D|}{N}

    Parameters
    ----------
    X : list
        A discrete array. Can contain digits 0-9 and/or
        single-character strings ("A", "a", "B", "b", etc.)..
    return_substrings : bool, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    float
        The estimated Lempel-Ziv complexity.
    set
        The dictionary (only returned if return_dictionary == True)

    References
    ----------

    Schartner, M. M., Carhart-Harris, R. L., Barrett, A. B., Seth, A. K., & Muthukumaraswamy, S. D. (2017). 
    Increased spontaneous MEG signal diversity for psychoactive doses of ketamine, LSD and psilocybin. 
    Scientific Reports, 7, 46421. 
    https://doi.org/10.1038/srep46421

    Blanc, J.-L., Schmidt, N., Bonnier, L., Pezard, L., & Lesne, A. (2008).
    Quantifying Neural Correlations Using Lempel-Ziv Complexity.
    Deuxième conférence française de Neurosciences Computationnelles,
    Marseille, France.
    https://hal.science/hal-00331599/document
    
    Zozor, S., Ravier, P., & Buttelli, O. (2005).
    On Lempel–Ziv complexity for multidimensional data analysis.
    Physica A: Statistical Mechanics and Its Applications, 345(1), 285–302.
    https://doi.org/10.1016/j.physa.2004.07.025
    """

    tuples: tuple = tuple((i,) for i in X)
    N: int = len(tuples)
    i = 0

    dictionary = {tuples[0]}  # Store unique substrings

    while i < N - 1:  # Scanning the string
        j = i + 1
        while j < N and tuples[i : j + 1] in dictionary:
            j += 1
        dictionary.add(tuples[i : j + 1])
        i = j  # Move index to next position

    c = len(dictionary)
    complexity = (c * np.log2(c)) / N  # The +1 is removed for consistency w/ Coutinho

    if return_dictionary is True:
        return complexity, dictionary
    else:
        return complexity


def lempel_ziv_mutual_information(X: list, Y: list) -> float:
    """
    Estimates the discrete mutual information rate for two channels X and Y with the Lempel-Ziv compression algorithm.
    This measure can be transiently negative, although in the limit it approximates the discrete information rate.

    .. math::
        I_{LZ}(X;Y) = LZ(X) + LZ(Y) - LZ(X,Y)

    Parameters
    ----------
    X : list
        A discrete list. Can contain digits 0-9 and/or
        single-character strings ("A", "a", "B", "b", etc.)..
    Y : list
        A discrete list. Can contain digits 0-9 and/or
        single-character strings ("A", "a", "B", "b", etc.)..

    Returns
    -------
    float
        The estimated mutual information rate.

    References
    ----------
    Zozor, S., Ravier, P., & Buttelli, O. (2005).
    On Lempel–Ziv complexity for multidimensional data analysis.
    Physica A: Statistical Mechanics and Its Applications, 345(1), 285–302.
    https://doi.org/10.1016/j.physa.2004.07.025

    Blanc, J.-L., Schmidt, N., Bonnier, L., Pezard, L., & Lesne, A. (2008).
    Quantifying Neural Correlations Using Lempel-Ziv Complexity.
    Deuxième conférence française de Neurosciences Computationnelles,
    Marseille, France.
    https://hal.science/hal-00331599/document

    """

    joint = list(zip(X, Y))

    return (
        lempel_ziv_complexity(X)
        + lempel_ziv_complexity(Y)
        - lempel_ziv_complexity(joint)
    )


def conditional_lempel_ziv_complexity(X: list, Y: list) -> float:
    """
    The conditional entropy rate estimated with the Lempel-Ziv algorithm.

    .. math::
        LZ(X|Y) = LZ(X,Y) - LZ(Y)

    Parameters
    ----------
    X : list
        A discrete list. Can contain digits 0-9 and/or
        single-character strings ("A", "a", "B", "b", etc.).
    Y : list
        A discrete list. Can contain digits 0-9 and/or
        single-character strings ("A", "a", "B", "b", etc.).

    Returns
    -------
    float
        The estimated conditional entropy rate.

    References
    ----------
    Zozor, S., Ravier, P., & Buttelli, O. (2005).
    On Lempel–Ziv complexity for multidimensional data analysis.
    Physica A: Statistical Mechanics and Its Applications, 345(1), 285–302.
    https://doi.org/10.1016/j.physa.2004.07.025

    """

    joint: list = list(zip(X, Y))

    return lempel_ziv_complexity(joint) - lempel_ziv_complexity(Y)


def lempel_ziv_total_correlation(data: np.ndarray) -> float:
    """
    A straightforward generalization of the mutual information rate given by
    Zozor et al., and Blanc et al.,

    .. math:: 
        TC_{LZ}(X) = \\sum_{i=1}^{N} LZ(X_i) - LZ(X)


    Parameters
    ----------
    data : np.ndarray
        A multi-dimensional discrete array, assumed to be in
        channels x time format.

    Returns
    -------
    float
        The estimated total correlation rate.

    """

    N0: int
    N1: int

    N0, N1 = data.shape

    sum_marginals: float = 0.0
    for i in range(N0):
        sum_marginals += lempel_ziv_complexity(data[i, :])

    joint = [tuple(data[:, i]) for i in range(N1)]

    return sum_marginals - lempel_ziv_complexity(joint)


def cross_lempel_ziv_complexity(X: list, Y: list) -> int:
    """
    Computes the Lempel-Ziv complexity of a string X
    using a pre-constructed dictionary optimized on Y. The relative
    Lempel-Ziv complexity is the extra patterns that appear in X but
    not in the compressed Y.

    This code uses the optimized dictionary for Y, rather than searching all
    substrings of Y. Using all possible substrings is very impractical for
    long time series. As a result, however, it is very sensitive to the
    particular temporal ordering of fluctuations in X and Y.

    Should only be used when X and Y were recorded at the same time, such
    as channels in an EEG/fMRI/MEG recording.

    MATHEMATICAL GARUNTEES GIVEN IN ZIV & MERHAV 1993 ARE NOT PRESERVED
    BY THIS METHOD. THIS FUNCTION REMAINS EXPERIMENTAL

    Parameters
    ----------
    X : np.ndarray
        A discrete array. Can contain digits 0-9 and/or
        single-character strings ("A", "a", "B", "b", etc.).
    Y : np.ndarray
        A discrete array. Can contain digits 0-9 and/or
        single-character strings ("A", "a", "B", "b", etc.).

    Returns
    -------
    complexity : int
        The Lempel-ziv relative complexity.

    References
    ----------
    Ziv, J., & Merhav, N. (1993).
    A Measure of Relative Entropy between Individual Sequences with
    Application to Universal Classification.
    Proceedings. IEEE International Symposium on Information Theory, 352–352.
    https://doi.org/10.1109/ISIT.1993.748668

    """
    X_tuples = tuple((i,) for i in X)
    N_X = len(X_tuples)

    # Step 1: Build LZ dictionary from Y (sequential parsing, no full substring storage)
    _, dictionary_Y = lempel_ziv_complexity(Y, return_dictionary=True)

    # Step 2: Parse X with dictionary_Y
    i, c = 0, 0
    while i < N_X - 1:
        j = i + 1
        while j <= N_X and X_tuples[i : j + 1] in dictionary_Y:
            j += 1
        dictionary_Y.add(X_tuples[i : j + 1])  # Store new substrings
        c += 1  # A new additions
        i = j  # Move forward

    complexity = c * np.log2(c) / N_X

    return complexity



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
sys.path.insert(0, os.path.abspath('../..'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Syntropy'
copyright = '2025, Thomas F. Varley'
author = 'Thomas F. Varley'
release = '0.0.1'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',        # Auto-generate docs from docstrings
    'sphinx.ext.napoleon',       # Support for NumPy/Google style docstrings
    'sphinx.ext.viewcode',       # Add links to source code
    'sphinx.ext.mathjax',        # Math support
    #'sphinx_autodoc_typehints',  # Type hints in docs
]

templates_path = ['_templates']
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']

# Napoleon settings (for docstring styles)
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

# Typehints settings
autodoc_typehints = "description"

autodoc_member_order = 'bysource'



from .shannon import (
    differential_entropy,
    mutual_information,
)

from .multivariate_mi import (
    total_correlation,
    higher_order_information
)

__all__ = [
    "differential_entropy",
    "mutual_information",
    "total_correlation",
    "higher_order_information"
]



from .shannon import (
    differential_entropy,
    local_differential_entropy, 
    conditional_entropy,
    local_conditional_entropy,
    mutual_information,
    local_mutual_information, 
    conditional_mutual_information,
    local_conditional_mutual_information,
    kullback_leibler_divergence,
    local_kullback_leibler_divergence,
)

from .multivariate_mi import (
    local_total_correlation,
    total_correlation,
    local_dual_total_correlation,
    dual_total_correlation,
    local_s_information,
    s_information,
    local_o_information,
    o_information,
    tse_complexity,
    description_complexity,
)

from .decompositions import (
    partial_entropy_decomposition,
    partial_information_decomposition,
    generalized_information_decomposition,
    representational_complexity,
)

from .temporal import (
    differential_entropy_rate,
    mutual_information_rate,
    total_correlation_rate,
    dual_total_correlation_rate,
    s_information_rate,
    o_information_rate
)

__all__ = [
    "differential_entropy",
    "local_differential_entropy", 
    "conditional_entropy",
    "local_conditional_entropy",
    "mutual_information",
    "local_mutual_information", 
    "conditional_mutual_information",
    "local_conditional_mutual_information",
    "kullback_leibler_divergence",
    "local_kullback_leibler_divergence",
    "local_total_correlation",
    "total_correlation",
    "local_dual_total_correlation",
    "dual_total_correlation",
    "local_s_information",
    "s_information",
    "local_o_information",
    "o_information",
    "tse_complexity",
    "description_complexity",
    "partial_entropy_decomposition",
    "partial_information_decomposition",
    "generalized_information_decomposition",
    "representational_complexity",
    "differential_entropy_rate",
    "mutual_information_rate",
    "total_correlation_rate",
    "dual_total_correlation_rate",
    "s_information_rate",
    "o_information_rate",
]



import numpy as np
import networkx as nx
from typing import Callable
from .utils import make_powerset, reduce_state

from ..lattices import LATTICE_2, LATTICE_3, LATTICE_4

BOTTOM_2: tuple = ((0,), (1,))
BOTTOM_3: tuple = ((0,), (1,), (2,))
BOTTOM_4: tuple = ((0,), (1,), (2,), (3,))

PATHS_2: dict = nx.shortest_paths.shortest_path_length(
    LATTICE_2, source=None, target=BOTTOM_2
)
LAYERS_2: dict = {
    val: {key for key in PATHS_2.keys() if PATHS_2[key] == val}
    for val in set(PATHS_2.values())
}

PATHS_3: dict = nx.shortest_paths.shortest_path_length(
    LATTICE_3, source=None, target=BOTTOM_3
)
LAYERS_3: dict = {
    val: {key for key in PATHS_3.keys() if PATHS_3[key] == val}
    for val in set(PATHS_3.values())
}

PATHS_4: dict = nx.shortest_paths.shortest_path_length(
    LATTICE_4, source=None, target=BOTTOM_4
)
LAYERS_4: dict = {
    val: {key for key in PATHS_4.keys() if PATHS_4[key] == val}
    for val in set(PATHS_4.values())
}

def local_precompute_sources(joint_distribution: dict[tuple, float]) -> dict:
    """
    A utility function that computes the local entropy of each subset of
    elements. This speeds up the computation using the hmin function
    considerably,

    Parameters
    ----------
    joint_distribution: dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    dict
        For each state, the joint entropy of every subset.

    """

    N: int = len(list(joint_distribution.keys())[0])

    sources: list = list(make_powerset(range(N)))
    sources.remove(())

    local_entropies: dict = {
        state: {source: {} for source in sources} for state in joint_distribution.keys()
    }

    state: tuple
    for state in local_entropies.keys():

        source: tuple
        for source in local_entropies[state]:

            # For a given state x, and a given source s, computes the total probability mass of the joint
            # distribution consistent with the state of the source.
            # E.g. if state = (0,0,0) and source = (0,1), probability_mass = P(X0=0 AND X1=0)
            probability_mass = sum(
                [
                    joint_distribution[key]
                    for key in joint_distribution.keys()
                    if reduce_state(key, source) == reduce_state(state, source)
                ]
            )

            if probability_mass > 0:
                local_entropies[state][source] = -np.log2(probability_mass)
            else:
                local_entropies[state][
                    source
                ] = 0  # Set log2(0) to 0 since impossible events contain no information.

    return local_entropies


def hmin_discrete_redundancy(
    atom: tuple, state: tuple, sources: dict, joint_distribution: dict[tuple, float]
) -> float:
    """
    For a collection of sources :math:`\\alpha=\\{a_1, a_2, \\ldots, a_k\\}`, computes
    the redundnat entropy shared by all sources as:


    :math:`h_{\\cap}^{min}(\\alpha) = \\min_{i}h(a_i)`

    See:
        Finn, C., & Lizier, J. T. (2020).
        Generalised Measures of Multivariate Information Content.
        Entropy, 22(2), Article 2.
        https://doi.org/10.3390/e22020216

    Parameters
    ----------
    atom : tuple
        The partial entropy atom.
    state : tuple
        The state of the system.
    sources : dict
        The pre-computed local entropies constructed by the
        precompute_local_entropies() function.
    joint_distribution: dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    float
        The redundant entropy shared by every source in the atom.

    """

    if joint_distribution[state] == 0:
        return 0
    else:
        return min(sources[state][source] for source in atom)


def imin_discrete_redundancy(
    atom: tuple,
    state: tuple,
    inputs: tuple,
    target: tuple,
    sources: dict,
    joint_distribution: dict[tuple, float],
) -> float:
    """
    For a collection of sources :math:`\\alpha = \\{a_{1}, a_{2}, \\ldots, a_{k}\\}` and a target :math:`t` the redundancy is defined as:

    :math:`i_{min}(\\alpha;t) = \\min_{i}h(a_i) - \\min_{i}h(a_i|t)`

    See:
        Finn, C., & Lizier, J. T. (2020).
        Generalised Measures of Multivariate Information Content.
        Entropy, 22(2), Article 2.
        https://doi.org/10.3390/e22020216


    Parameters
    ----------
    atom : tuple
        The partial information atom.
    state : tuple
        The joint-state of the systems.
    inputs : tuple
        The indices of the input variables.
    target : tuple
        The indices of the target variable. May be multivariate.
    sources : dict
        The pre-computed local entropies
        for the joint state of the source and the target.
    joint_distribution: dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    float
        The local redundant mutual information.

    """

    i_plus: float = np.inf
    i_minus: float = np.inf

    h_target: float = sources[state][target]

    for source in atom:

        input_source: tuple = tuple(inputs[x] for x in source)
        h_input_source: float = sources[state][input_source]

        if i_plus > h_input_source:
            i_plus = h_input_source

        joint_source: tuple = tuple(sorted(input_source + target))

        h_conditional: float = sources[state][joint_source] - h_target

        if i_minus > h_conditional:
            i_minus = h_conditional

    return i_plus - i_minus


def hsx_discrete_redundancy(
    atom: tuple, state: tuple, joint_distribution: dict[tuple, float]
) -> float:
    """
    Computes the redundant entropy shared by a set of sources using the :math:`h_{sx}` function.

    For a collection of sources :math:`\\alpha = \\{a_{1}, a_{2}, \\ldots, a_{k}\\}`,
    the redundancy is defined as

    :math:`h^{sx}_{\cap}(\\alpha) = -\\log_{2} P(a_{1} \\cup a_{2} \\cup \\ldots \\cup a_{k})`.

    See:
        Varley, T. F., Pope, M., Maria Grazia, P., Joshua, F., & Sporns, O. (2023).
        Partial entropy decomposition reveals higher-order information structures in human brain activity.
        Proceedings of the National Academy of Sciences, 120(30), e2300888120.
        https://doi.org/10.1073/pnas.2300888120

    Parameters
    ----------
    atom : tuple
        The partial entropy atom.
    state : tuple
        The state of the system.
    joint_distribution: dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    float
        The redundant entropy shared by every source in the atom..

    """
    if joint_distribution[state] == 0:
        return 0.0
    else:
        state_set: set = set()

        source: tuple
        for source in atom:

            state_source: tuple = reduce_state(state, source)
            state_set.update(
                {
                    key
                    for key in joint_distribution.keys()
                    if reduce_state(key, source) == state_source
                }
            )

        redundant_entropy: float = -np.log2(
            sum(joint_distribution[s] for s in state_set)
        )

        return redundant_entropy


def isx_discrete_redundancy(
    atom: tuple,
    state: tuple,
    inputs: tuple,
    target: tuple,
    joint_distribution: dict[tuple, float],
) -> float:
    """
    For a collection of sources :math:`\\alpha = \\{a_{1}, a_{2}, \\ldots, a_{k}\\}` and a target :math:`t` the redundancy is defined as:

    :math:`i_{sx}(\\alpha;t) = \\log_{2}\\frac{P(t) - P(t \\cap (\\bar{a}_1 \\cap \\ldots \\cap \\bar{a}_k)}{1 - P(\\bar{a}_1 \\cap \\ldots \\cap \\bar{a}_{k})} - \\log_2 P(t)`

    Parameters
    ----------
    atom : tuple
        The partial information atom.
    state : tuple
        The joint-state of the systems.
    inputs : tuple
        The indices of the input variables.
    target : tuple
        The indices of the target variable. May be multivariate.
    sources : dict
        The pre-computed local entropies
        for the joint state of the source and the target.
    joint_distribution: dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.

    Returns
    -------
    float
        The redundant mutual informations.

    """

    # i_sx^+
    # -log2(1 / a1 U a2 U ... U ak)

    state_set: set = set()
    source: tuple
    for source in atom:

        input_source = tuple(inputs[x] for x in source)
        state_source: tuple = reduce_state(state, input_source)
        state_set.update(
            {
                key
                for key in joint_distribution.keys()
                if reduce_state(key, input_source) == state_source
            }
        )

    i_plus = np.log2(1 / sum(joint_distribution[x] for x in state_set))

    # i_sx^-
    # -log2( P(t) / t ^ (a1 U a2 U ... U ak))
    target_state: tuple = reduce_state(state, target)
    p_target: float = sum(
        [
            joint_distribution[key]
            for key in joint_distribution.keys()
            if reduce_state(key, target) == target_state
        ]
    )

    state_set: set = set()
    for source in atom:

        input_source = tuple(inputs[x] for x in source)
        state_source: tuple = reduce_state(state, input_source)
        state_set.update(
            {
                key
                for key in joint_distribution.keys()
                if reduce_state(key, input_source) == state_source
            }
        )

    state_set = {key for key in state_set if reduce_state(key, target) == target_state}

    i_minus = np.log2(p_target / sum([joint_distribution[x] for x in state_set]))

    return i_plus - i_minus


def mobius_inversion(
    decomposition: str,
    joint_distribution: dict[tuple, float],
    redundancy: str,
    inputs: tuple = (None,),
    target: tuple = (None,),
) -> tuple[dict, dict]:
    """
    Computes the Mobius inversion on the antichain lattice.

    Parameters
    ----------
    decomposition : str
        Which decomposition to use:
            For partial information decomposition use "pid".
            For partial entropy decomposition use "ped".
    joint_distribution: dict[tuple, float]
        The joint probability distribution.
        Keys are tuples corresponding to the state of each element.
        The valules are the probabilities.
    redundancy : str
        The redundancy function.
    inputs : tuple, optional
        The indicies of the inputs. The default is (None,)
    target : tuple, optional
        The (potentially multivariate) indices of the target.
        The default is (None,).

    Returns
    -------
    (dict, dict)
        The pointwise and average partial information atom dictionaries.

    """


    decomposition_lower = decomposition.lower()
    assert decomposition_lower in {
        "pid",
        "ped",
    }, "You must specify a decomposition: PID or PED."

    if decomposition_lower == "pid":
        assert redundancy in {
            "isx",
            "imin",
        }, "The implemented redundancy functions are 'imin' and 'hsx'."
        assert target != (None,), "You must specify a target."
        if redundancy == "isx":
            redundancy_function = isx_discrete_redundancy
        elif redundancy == "imin":
            redundancy_function = imin_discrete_redundancy
        total_str: str = "total_information"
        partial_str: str = "pi"
    elif decomposition_lower == "ped":
        assert redundancy in {
            "hsx",
            "hmin",
        }, "The implemented redundancy functions are 'hxs' and 'hmin'."
        if redundancy == "hsx":
            redundancy_function = hsx_discrete_redundancy
        elif redundancy == "hmin":
            redundancy_function = hmin_discrete_redundancy
        total_str: str = "total_entropy"
        partial_str: str = "pe"

    if inputs == (None,):  # If no inputs are specified, all elements are inputs
        N = {len(x) for x in joint_distribution.keys()}.pop()
    else:
        N = len(inputs)

    if N == 2:
        lattice, bottom, layers = LATTICE_2, BOTTOM_2, LAYERS_2
    elif N == 3:  #    |
        lattice, bottom, layers = LATTICE_3, BOTTOM_3, LAYERS_3
    else:  #                      |
        lattice, bottom, layers = LATTICE_4, BOTTOM_4, LAYERS_4

    # Pre-computing all the local h_{\partial}(source) values
    # this saves a lot of time.

    if "min" in redundancy:
        sources: dict = local_precompute_sources(joint_distribution)

    ptw: dict = {state: dict() for state in joint_distribution.keys()}

    for state in joint_distribution.keys():
        for layer in layers:
            for atom in layers[layer]:
                if decomposition_lower == "pid":

                    if redundancy == "imin":
                        args = {
                            "atom": atom,
                            "state": state,
                            "inputs": inputs,
                            "target": target,
                            "sources": sources,
                            "joint_distribution": joint_distribution,
                        }
                    elif redundancy == "isx":
                        args = {
                            "atom": atom,
                            "state": state,
                            "inputs": inputs,
                            "target": target,
                            "joint_distribution": joint_distribution,
                        }

                    lattice.nodes[atom][total_str] = redundancy_function(**args)
                elif decomposition_lower == "ped":

                    if redundancy == "hmin":
                        args = {
                            "atom": atom,
                            "state": state,
                            "sources": sources,
                            "joint_distribution": joint_distribution,
                        }
                    elif redundancy == "hsx":
                        args = {
                            "atom": atom,
                            "state": state,
                            "joint_distribution": joint_distribution,
                        }

                    lattice.nodes[atom][total_str] = redundancy_function(**args)

                if atom == bottom:
                    lattice.nodes[atom][partial_str] = lattice.nodes[atom][total_str]
                else:
                    lattice.nodes[atom][partial_str] = lattice.nodes[atom][
                        total_str
                    ] - sum(
                        [
                            lattice.nodes[d][partial_str]
                            for d in lattice.nodes[atom]["descendants"]
                        ]
                    )

        local_ptw: dict = {
            node: lattice.nodes[node][partial_str] for node in lattice.nodes
        }
        ptw[state] = local_ptw

    avg = {}
    for node in lattice.nodes:
        avg[node] = sum(
            [joint_distribution[state] * ptw[state][node] for state in ptw.keys()]
        )

    return ptw, avg
def partial_information_decomposition(
    redundancy: str,
    inputs: tuple,
    target: tuple,
    joint_distribution: dict[tuple[int, ...], float],
) -> (dict, dict):
    """
    Computes the partial information decomposition for up to four input variables
    onto one (potentially joint) target variable.

    The available redundancy functions are :math:`i_{\\min}` from Finn and Lizier and :math:`i_{sx}` from Makkeh et al..

    .. math:: 

        i_{\\min}(\\alpha;t) &= \\min h(\\alpha_i) - \min h(\\alpha_i|t) \\\\
           i_{sx}(\\alpha;t) &= \\log\\frac{P(t)-P(t\\cap(\\alpha_1\\cup ...\\cup\\alpha_k))}{1-P(\\bar\\alpha_1\\cap ...\\cap\\alpha_N)}
    
    Parameters
    ----------
    redundancy : str
        The redundancy function.
        "imin" for the Finn and Lizier measure
        "isx" for the Makkeh et al., measure
    inputs : tuple
        The set of up to four input elements.
    target : tuple
        The set of target elements.
        If len(target) > 1, then the target is the joint
        state of all elements
    joint_distribution : dict[tuple,float]
        The joint distribution object.

    Returns
    -------
    ptw : dict
        A dictionary of dictionaries,
        The outer dictionary has one key for each joint state.
        Each inner dictionary is the lookup of partial information atoms.
    avg : dict
        The expected value for each partial information atom.

    References
    ----------
    Williams, P. L., & Beer, R. D. (2010). 
    Nonnegative Decomposition of Multivariate Information. 
    arXiv:1004.2515 [Math-Ph, Physics:Physics, q-Bio]. 
    http://arxiv.org/abs/1004.2515

    Finn, C., & Lizier, J. T. (2018).
    Pointwise Partial Information Decomposition Using
    the Specificity and Ambiguity Lattices.
    Entropy, 20(4), Article 4.
    https://doi.org/10.3390/e20040297

    Makkeh, A., Gutknecht, A. J., & Wibral, M. (2021).
    Introducing a differentiable measure of pointwise
    shared information.
    Physical Review E, 103(3), 032149.
    https://doi.org/10.1103/PhysRevE.103.032149
    
    """

    ptw, avg = mobius_inversion(
        decomposition="pid",
        joint_distribution=joint_distribution,
        inputs=inputs,
        redundancy=redundancy,
        target=target,
    )

    return ptw, avg


def partial_entropy_decomposition(
    redundancy: str, joint_distribution: dict[tuple[int, ...], float]
) -> (dict, dict):
    """
    Computes the partial entropy decomposition of a joint distribution
    with up to four elements.

    The available redundancy functions are h_min from Finn and Lizier
    and h_sx from Varley et al.,
    
    .. math:: 
        h_{\\min}(\\alpha) &= \\min(\\alpha_i) \\\\
           h_{sx}(\\alpha) &= \\log\\frac{1}{P(\\alpha_1\\cup ... \\cup\\alpha_N)}

    Parameters
    ----------
    redundancy : str
        The redundnacy function to use.
        "hmin" for the measure from Finn and Lizier,
        "hsx" for the measure from Varley et al.,
    joint_distribution : dict[tuple,float]
        The joint distribution object.

    Returns
    -------
    ptw : dict[tuple, dict]
        A dictionary of dictionaries,
        The outer dictionary has one key for each joint state.
        Each inner dictionary is the lookup of partial entropy atoms.
    avg : dict[tuple, float]
        The expected value for each partial entropy atom.

    References
    ----------
    Finn, C., & Lizier, J. T. (2020).
    Generalised Measures of Multivariate Information Content.
    Entropy, 22(2), Article 2.
    https://doi.org/10.3390/e22020216

    Varley, T. F., Pope, M., Maria Grazia, P., Joshua, F., & Sporns, O. (2023).
    Partial entropy decomposition reveals higher-order
    information structures in human brain activity.
    Proceedings of the National Academy of Sciences,
    120(30), e2300888120.
    https://doi.org/10.1073/pnas.2300888120

    """

    ptw, avg = mobius_inversion(
        decomposition="ped",
        joint_distribution=joint_distribution,
        redundancy=redundancy,
    )

    return ptw, avg


def generalized_information_decomposition(
    redundancy: str,
    posterior_distribution: dict[tuple[int, ...], float],
    prior_distribution: dict[tuple[int, ...], float],
) -> (dict, dict):
    """
    Computes the generalized information decomposition from Varley et al.
    The GID is a decomposition of the Kullback-Leibler divergence of a
    posterior distribution from a prior distribution.

    Available redundnacy functions are "hmin" and "hsx". See
    the documentation for the partial_entropy_decomposition() function
    for details.

    Parameters
    ----------
    redundancy : str
        The localizable redundancy function.
        Options are: hmin and hsx.
    posterior_distribution : dict[tuple, float]
        The posterior distribution.
        The support set of this distribution must be a subset of
        the supppirt set of the prior distribution.
    prior_distribution : dict[tuple, float]
        The prior distribution.

    Returns
    -------
    ptw : dict[tuple, dict]
        A dict of dicts.
        The local Kullback-Leibler divergence for each atom for each state.
    avg : dict[tuple, float]
        The average Kullback-Leibler divergence for each atom.

    References
    ----------
    Varley, T. F. (2024).
    Generalized decomposition of multivariate information.
    PLOS ONE, 19(2), e0297128.
    https://doi.org/10.1371/journal.pone.0297128

    """
    assert set(prior_distribution.keys()).issuperset(
        set(posterior_distribution.keys())
    ), (
        "The support set of the prior must be a superset of the support set of the posterior."
    )
    assert redundancy.lower() in {
        "hmin",
        "hsx",
    }, "The supported redundancy functions are hmin and hsx."

    ptw_prior, _ = mobius_inversion(
        decomposition="ped",
        joint_distribution=prior_distribution,
        redundancy=redundancy,
    )

    ptw_posterior, _ = mobius_inversion(
        decomposition="ped",
        joint_distribution=posterior_distribution,
        redundancy=redundancy,
    )

    nodes = list(ptw_prior[list(prior_distribution.keys())[0]].keys())

    ptw = {
        state: {
            node: ptw_prior[state][node] - ptw_posterior[state][node] for node in nodes
        }
        for state in posterior_distribution.keys()
    }

    avg = {}
    for node in nodes:
        avg[node] = sum(
            [posterior_distribution[state] * ptw[state][node] for state in ptw.keys()]
        )

    return ptw, avg


def representational_complexity(avg: dict, comparator: Callable = min) -> float:
    """
    Computes the representational complexity of a given partial information or entropy lattice.
    The representational complexity is a measure of how
    much partial information atoms of a given degree of synergy
    contribute to the overall mutual information or entropy.


    Parameters
    ----------
    avg : dict[tuple, float]
        The dictionary of partial information/entropy atoms.
        Returned from any of the above functions.
    comparator : function, optional
        Whether to consider the minimum complexity of an atom.
        or the maximum complexity of an atom.
        Options are: min, max, np.min, np.max.
        The default is min, following the original work
        by Ehrlich et al.,.

    Returns
    -------
    float
        The representational complexity.

    References
    ----------
    Ehrlich, D. A., Schneider, A. C., Priesemann, V., Wibral, M., & Makkeh, A. (2023).
    A Measure of the Complexity of Neural Representations
    based on Partial Information Decomposition.
    Transactions on Machine Learning Research.
    https://openreview.net/forum?id=R8TU3pfzFr

    """

    assert comparator in (min, max, np.min, np.max), "The comparator must be min or max"

    rc: float = 0.0

    for atom in avg.keys():
        rc += avg[atom] * comparator(len(source) for source in atom)

    return rc / sum(avg.values())



import pytest 

from syntropy.discrete.optimization import constrained_maximum_entropy_distributions
from syntropy.discrete.distributions import (GIANT_BIT, XOR_DIST, SUMMED_DICE, 
                                          RANDOM_DIST_2, RANDOM_DIST_3, RANDOM_DIST_4)
from syntropy.discrete import multivariate_mi as mi 
from syntropy.discrete import shannon
from syntropy.discrete.decompositions import (partial_entropy_decomposition as ped, 
                                           partial_information_decomposition as pid, 
                                           generalized_information_decomposition as gid)


def test_total_correlation():
    
    assert mi.total_correlation(GIANT_BIT)[1] == pytest.approx(2.0)
    assert mi.total_correlation(XOR_DIST)[1] == pytest.approx(1.0)
    
    ptw, avg = mi.total_correlation(SUMMED_DICE)
    
    assert avg == pytest.approx(3.2744019192887714)
    assert sum(SUMMED_DICE[key]*ptw[key] for key in ptw.keys()) == pytest.approx(avg)


def test_dual_total_correlation():
    
    assert mi.dual_total_correlation(GIANT_BIT)[1] == pytest.approx(1.0)
    assert mi.dual_total_correlation(XOR_DIST)[1] == pytest.approx(2.0)
    
    ptw, avg = mi.dual_total_correlation(SUMMED_DICE)
    
    assert avg == pytest.approx(5.169925001442313)
    assert sum(SUMMED_DICE[key]*ptw[key] for key in ptw.keys()) == pytest.approx(avg)


def test_s_information():
    
    assert mi.s_information(GIANT_BIT)[1] == pytest.approx(3.0)
    assert mi.s_information(XOR_DIST)[1] == pytest.approx(3.0)
    
    ptw, avg = mi.s_information(SUMMED_DICE)
    
    assert avg == pytest.approx(8.444326920731084)
    assert sum(SUMMED_DICE[key]*ptw[key] for key in ptw.keys()) == pytest.approx(avg)

    
def test_o_information():
    
    assert mi.o_information(GIANT_BIT)[1] == pytest.approx(1.0)
    assert mi.o_information(XOR_DIST)[1] == pytest.approx(-1.0)
    
    ptw, avg = mi.o_information(SUMMED_DICE)
    
    assert avg == pytest.approx(-1.8955230821535416)
    assert sum(SUMMED_DICE[key]*ptw[key] for key in ptw.keys()) == pytest.approx(avg)


def test_mutual_information():
    
    ptw_tc, avg_tc = mi.total_correlation(RANDOM_DIST_2)
    ptw_mi, avg_mi = shannon.mutual_information((0,),(1,), RANDOM_DIST_2)
    
    assert avg_tc == pytest.approx(avg_mi)
    
    ptw_mi = {sum(key,()) : ptw_mi[key] for key in ptw_mi.keys()}
    assert len(ptw_mi) == len(ptw_tc)
    assert False not in {ptw_mi[key] == pytest.approx(ptw_tc[key]) for key in ptw_mi.keys()}

    _, avg_s = mi.s_information(RANDOM_DIST_3)
    total = 0
    for i in range(3):
        total += shannon.mutual_information((i,), tuple(j for j in range(3) if j != i),
                                            RANDOM_DIST_3)[1]
    assert total == pytest.approx(avg_s)
    

def test_conditional_mutual_information():

    idxs_x = (0,)
    idxs_y = (1,)
    idxs_z = (2,)
    
    ptw, avg = shannon.conditional_mutual_information(
        idxs_x, idxs_y, idxs_z, XOR_DIST)
    assert avg == pytest.approx(1.0)
    
    ptw, avg = ptw, avg = shannon.conditional_mutual_information(
        idxs_x, idxs_y, idxs_z, GIANT_BIT)
    assert avg == pytest.approx(0.0)
    
    ptw, avg = ptw, avg = shannon.conditional_mutual_information(
        idxs_x, idxs_y, idxs_z, SUMMED_DICE)
    assert avg == pytest.approx(1.8955230821535425)
    
    test_avg = 0.0
    for key in ptw.keys():
        state = sum(key, ())
        test_avg += ptw[key]*SUMMED_DICE[state]
    assert test_avg == pytest.approx(avg)
    

def test_co_information():
    
    ptw_o, avg_o = mi.o_information(RANDOM_DIST_3)
    ptw_co, avg_co = mi.co_information(RANDOM_DIST_3)
    
    assert avg_o == pytest.approx(avg_co)
    assert avg_co == pytest.approx(-0.07853256580219448)


def test_tse():
    
    assert mi.tse_complexity(XOR_DIST, 10) == pytest.approx(1.0)
    assert mi.tse_complexity(GIANT_BIT, 10) == pytest.approx(1.0)
    

def test_ped():
    
    ptw, avg = ped(redundancy = "hmin", joint_distribution = XOR_DIST)
    assert avg[((0,),(1,),(2,))] == pytest.approx(1.0)
    assert avg[((0,1),(0,2),(1,2))] == pytest.approx(1.0)
    assert avg[((0,1,2),)] == pytest.approx(0.0)
    assert sum(avg.values()) == shannon.shannon_entropy(XOR_DIST)[1]
    
    ptw, avg = ped(redundancy = "hsx", joint_distribution = XOR_DIST)
    assert avg[((0,),(1,),(2,))] == pytest.approx(0.0)
    assert avg[((0,1,2),)] == pytest.approx(0.0)
    assert avg[((0,1),(0,2),(1,2))] == pytest.approx(0.24511249783653144)
    assert avg[((0,),(1,2))] == pytest.approx(0.16992500144231237)
    assert sum(avg.values()) == shannon.shannon_entropy(XOR_DIST)[1]


def test_pid():
    
    ptw, avg = pid(redundancy = "imin", joint_distribution = XOR_DIST,
                   inputs=(0,1), target=(2,))
    assert avg[((0,1),)] == pytest.approx(1.0)
    assert avg[((0,),(1,),)] == pytest.approx(0.0)
    assert avg[((0,),)] == pytest.approx(0.0)
    assert avg[((1,),)] == pytest.approx(0.0)
    assert sum(avg.values()) == pytest.approx(shannon.mutual_information((0,1), (2,), XOR_DIST)[1])
    
    ptw, avg = pid(redundancy = "isx", joint_distribution = XOR_DIST,
                   inputs=(0,1), target=(2,))
    assert avg[((0,1),)] == pytest.approx(0.4150374992788438)
    assert avg[((0,),(1,),)] == pytest.approx(-0.5849625007211562)
    assert avg[((0,),)] == pytest.approx(0.5849625007211563)
    assert avg[((1,),)] == pytest.approx(0.5849625007211563)
    assert sum(avg.values()) == pytest.approx(shannon.mutual_information((0,1), (2,), XOR_DIST)[1])


def test_gid():
    
    maxent = constrained_maximum_entropy_distributions(XOR_DIST, order=1)

    ptw, avg = gid(redundancy="hmin", 
                   posterior_distribution=XOR_DIST,
                   prior_distribution=maxent,
                   )
    
    assert avg[((0,1,2),)] == pytest.approx(1.0)
    assert sum(avg.values()) == pytest.approx(shannon.kullback_leibler_divergence(XOR_DIST, maxent)[1])

def test_connected_information():
    
    profile = mi.connected_information(RANDOM_DIST_4)
    tc = mi.total_correlation(RANDOM_DIST_4)[1]

    assert sum(profile) == pytest.approx(tc)



import torch
import torch.optim as optim
from torch.utils.data import DataLoader, TensorDataset

import nflows
from nflows.distributions.normal import StandardNormal
from nflows.flows.base import Flow
from nflows.transforms.autoregressive import MaskedAffineAutoregressiveTransform
from nflows.transforms.base import CompositeTransform
from nflows.transforms.permutations import RandomPermutation


def initialize_flow(
    dim: int,
    dim_context: int = 0,
    num_layers: int = 5,
    hidden_features: int = 64,
    dropout_probability: float = 0.1,
) -> nflows.flows.base.Flow:
    """
    Initializes a new normalizing flow network.

    Parameters
    ----------
    dim : int
        The number of input dimensions.
    dim_context : int
        The number of conditioning dimension.
        The default is zero.
    num_layers : int
        The number of hidden layers.
        The default is 5 layers.
    hidden_features : int
        The number of neurons in each hidden layer.
        The default is 64 neurons.
    dropout_probability : int
        The probability of a neuron dropping out.

    Returns
    -------
        nflows.flows.base.Flow
    """
    transforms: list = []

    for _ in range(num_layers):
        transforms.append(
            MaskedAffineAutoregressiveTransform(
                features=dim,
                hidden_features=hidden_features,
                context_features=dim_context,
                dropout_probability=dropout_probability,
            )
        )
        transforms.append(RandomPermutation(features=dim))

    transform = CompositeTransform(transforms)
    base_dist = StandardNormal(shape=[dim])

    return Flow(transform, base_dist)


def train_flow(
    flow: nflows.flows.base.Flow,
    data: torch.Tensor,
    context: None | torch.Tensor = None,
    batch_size: int = 256,
    lr: float = 1e-4,
    num_epochs: int = 100,
    weight_decay: float = 1e-5,
    convergence_threshold: float = 0.0,
    alpha: float = 0.1,
    verbose: bool = False,
) -> nflows.flows.base.Flow:
    """
    Trains a normalizing flow network to approximate the maximally likely distribution to have generated the given data.

    Parameters
    ----------
    flow : nflows.flows.base.Flow
        An untrained normalizing flow network.
    data : torch.Tensor
        The training data, in samples x features format.
    context : None | torch.Tensor
        Conditioning random variables.
        Default is None.
    batch_size : int
        The size of each batch.
        The default value is 256.
    lr : float
        The learning rate.
        The default value is 1e-4
    num_epochs : int
        The number of training epochs.
        The default value is 100,
    weight_decay: float
        The rate at which parameter weights decay. A regularizer to reduce over-fitting.
        The default is 1e-5.
    convergence_threshold : float
        The value of the coefficient of variation below which the training terminates.
        The default value is 0.0, which means the training will not stop before num_epochs is hit.
    alpha : float
        How quickly the exponentially weighted moving standard deviation downweights older data.
        The default is 0.2.
    verbose : bool
        Whether to print the loss for each epoch throughout training.
        The default is False.


    Returns
    -------

    nflows.flows.base.Flow
    """

    if context is None:
        dataset = TensorDataset(data)
    elif context is not None:
        dataset = TensorDataset(data, context)

    loader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

    optimizer = optim.Adam(flow.parameters(), lr=lr, weight_decay=weight_decay)

    flow.train()

    running_mean: float = 0.0
    running_var: float = 0.0

    for epoch in range(num_epochs):
        epoch_nll_sum: float = 0

        for batch in loader:
            optimizer.zero_grad()

            if context is None:
                x = batch[0]
                batch_nll = -flow.log_prob(x).mean()
            else:
                x, c = batch
                batch_nll = -flow.log_prob(x, context=c).mean()
            batch_nll.backward()

            optimizer.step()

            epoch_nll_sum += batch_nll.item() * x.shape[0]

        epoch_avg_nll: float = epoch_nll_sum / data.shape[0]

        delta = epoch_avg_nll - running_mean
        running_mean += alpha * delta
        running_var = (1 - alpha) * (running_var + alpha * delta**2)
        running_std = running_var ** (1 / 2)
        coef_var = running_std / (running_mean + 1e-8)

        if verbose is True:
            print(
                f"Epoch {epoch + 1}, NLL: {epoch_avg_nll:.3f}, Coef. var: {coef_var:.3f}"
            )

        if coef_var < convergence_threshold:
            return flow

    return flow


def evaluate_flow(
    flow: nflows.flows.base.Flow,
    data: torch.Tensor,
    context: None | torch.Tensor = None,
) -> tuple[float, float]:
    """
    Evaluates a trainied normalizing flow network on the given data.

    Parameters
    ----------
    flow : nflows.flows.base.Flow
        A trained normalilzing flow network.
    data : torch.Tensor
        The testing data in samples x features format.
    context : None | torch.Tensor


    Returns
    -------
    tuple[float, float]
        The average entropy (in nat)
        The standard error of the estimate.

    """

    flow.eval()

    with torch.no_grad():
        if context is None:
            log_probs = flow.log_prob(data)
        else:
            log_probs = flow.log_prob(data, context=context)
        h = -log_probs.mean().item()

    return -log_probs, h



import numpy as np
import scipy.stats as stats
from numpy.typing import NDArray
from .utils import check_cov

H_SINGLE: float = np.log(np.sqrt(2.0 * np.pi * np.e))
LN_TWO_PI_E: float = np.log(2.0 * np.pi * np.e)
TWO_PI: float = 2.0 * np.pi
SQRT_TWO_PI: float = np.sqrt(TWO_PI)

COV_NULL: NDArray[np.floating] = np.array([[-1.0]])


def differential_entropy(
    cov: NDArray[np.floating], idxs: tuple[int, ...] = (-1,)
) -> float:
    """
    Computes the expected differential entropy of a multivariate
    distribution parameterized by a covariance matrix using a
    Gaussian estimator.

    The differential entropy is given by:

    .. math::

        H(X) = \\int dx P(x)\\log P(x)

    And if :math:`X` is drawn from a k-dimensional Gaussian, it is equal to

    .. math::

        H(X) = \\frac{k}{2}\\log 2\\pi\\textnormal{e} + \\frac{1}{2}\\log|\\Sigma|

    Where :math:`|\\Sigma|` is the determinant of the covariance matrix.

    Parameters
    ----------
    cov : NDArray[np.floating]
        The covariance matrix that defines the distribution.
    idxs : tuple[int, ...], optional
        The specific subset of variables to compute the total correlation of.
        Defaults to computing the TC of the entire covariance matrix.

    Returns
    -------
    float

    """

    if idxs[0] == -1:
        return stats.multivariate_normal(cov=cov, allow_singular=True).entropy()
    else:
        if len(idxs) == 1:
            return H_SINGLE
        else:
            return stats.multivariate_normal(
                cov=cov[np.ix_(idxs, idxs)], allow_singular=True
            ).entropy()


def local_differential_entropy(
    data: NDArray[np.floating], cov: NDArray[np.floating] = COV_NULL
) -> NDArray[np.floating]:
    """
    Computes the framewise differential entropy for a set of variables.

    .. math::
        h(x) = -\\log P(x)

    For data drawn from a k-dimensional Gaussian

    .. math::
        P(x) = (2\\pi)^{-k/2}|\\Sigma|^{-1/2}\\textnormal{e}^{\\frac{-(x - \\mu)^\\mathrm{T} \\Sigma^{-1}(x - \\mu)}{2}}

    Parameters
    ----------
    data : NDArray[np.floating]
        The data in channels x samples format.
    cov : NDArray[np.floating], optional
        The covariance matrix that defines the distribution.
        If none is provided, it is computed from the data object.

    Returns
    -------
    NDArray[np.floating]
        The series of pointwise entropies.

    """
    N: int = data.shape[0]

    cov_: NDArray[np.floating] = check_cov(cov, data)

    if N == 1:
        return -stats.norm.logpdf(
            x=data, loc=data.mean(), scale=data.std()
        )
    else:
        return -(
            stats.multivariate_normal.logpdf(
                x=data.T, mean=data.mean(axis=-1), cov=cov_, allow_singular=True
            )
        )


def conditional_entropy(
    idxs_x: tuple[int, ...],
    idxs_y: tuple[int, ...],
    cov: NDArray[np.floating],
) -> float:
    """
    Computes the conditional entropy of X given Y using Gaussian estimation.
    
    .. math::
        H(X|Y) = H(X,Y) - H(Y)

    Parameters
    ----------
    idxs_x : tuple
        The indices of the variables to compute the conditional entropy on.
    idxs_y : tuple
        The indices of the conditioning set.
    cov : NDArray[np.floating], optional
        The covariance matrix that defines the distribution.
        If none is provided, it is computed from the data object.

    Returns
    -------
    float

    """

    joint: tuple[int, ...] = idxs_x + idxs_y

    h_joint: float = differential_entropy(cov, joint)
    h_y: float = differential_entropy(cov, idxs_y)

    return h_joint - h_y


def local_conditional_entropy(
    idxs_x: tuple[int, ...],
    idxs_y: tuple[int, ...],
    data: NDArray[np.floating],
    cov: NDArray[np.floating] = COV_NULL,
) -> NDArray[np.floating]:
    """
    Computes the local condition entropy for every sample in data using Gaussian estimation.

    .. math::
    
        h(x|y) = h(x,y) - h(y)

    Parameters
    ----------
    idxs_x : tuple
        The indices of the variables to compute the conditional entropy on.
    idxs_y : tuple
        The indices of the conditioning set.
    data : NDArray[np.floating]
        The data in channels x samples format.
    cov : NDArray[np.floating], optional
        The covariance matrix that defines the distribution.
        If none is provided, it is computed from the data object.

    Returns
    -------
    NDArray[np.floating]

    """

    cov_: NDArray[np.floating] = check_cov(cov, data)

    joint: tuple[int, ...] = idxs_x + idxs_y

    h_y = local_differential_entropy(data[idxs_y, :], cov_[np.ix_(idxs_y, idxs_y)])
    h_joint = local_differential_entropy(data[joint, :], cov_[np.ix_(joint, joint)])

    return h_joint - h_y


def mutual_information(
    idxs_x: tuple[int, ...], idxs_y: tuple[int, ...], cov: NDArray[np.floating]
) -> float:
    r"""
    Computes the mutual information between two (potentially multivariate) sets of elements.

    .. math::
        I(X;Y) &= H(X) + H(Y) - H(X,Y) \\
               &= H(X) - H(X|Y) \\
               &= H(Y) - H(Y|X) \\
               &= H(X,Y) - H(X|Y) - H(Y|X)

    For Gaussian random variables:

    .. math::
        I(X;Y) = \frac{1}{2}\log\frac{|\Sigma_{X}||\Sigma_{Y}|}{|\Sigma_{XY}|}

    In the particular case where :math:`X` and :math:`Y` are univariate, the mutual information can be computed directly from the Pearson correlation coefficient :math:`r`:

    .. math::
        I(X;Y) = \frac{-\log(1-r^{2})}{2}

    Parameters
    ----------
    idxs_x : tuple
        The indices of the source variables. Can be multivariate.
    idxs_y : tuple
        The indices of the idxs_y variable. Can be multivariate.
    cov : NDArray[np.floating]
        The covariance matrix that defines the distribution.

    Returns
    -------
    float

    """
    joint: tuple[int, ...] = idxs_x + idxs_y

    cov_idxs_x: NDArray[np.floating] = cov[np.ix_(idxs_x, idxs_x)]
    cov_idxs_y: NDArray[np.floating] = cov[np.ix_(idxs_y, idxs_y)]
    cov_joint: NDArray[np.floating] = cov[np.ix_(joint, joint)]

    det_idxs_x: float = 0.0
    if len(idxs_x) == 1:
        det_idxs_x += np.linalg.det([[cov_idxs_x]])
    else:
        det_idxs_x += np.linalg.det(cov_idxs_x)

    det_idxs_y: float = 0.0
    if len(idxs_y) == 1:
        det_idxs_y += np.linalg.det([[cov_idxs_y]])
    else:
        det_idxs_y += np.linalg.det(cov_idxs_y)

    det_joint: float = np.linalg.det(cov_joint)

    dets: float = (det_idxs_x * det_idxs_y) / det_joint

    return (np.log(dets) / 2).item()


def local_mutual_information(
    idxs_x: tuple[int, ...],
    idxs_y: tuple[int, ...],
    data: NDArray[np.floating],
    cov: NDArray[np.floating] = COV_NULL,
) -> NDArray[np.floating]:
    """
    Computes the local mutual information between X and Y for every sample in data using Gaussian estimation.
    Note that the local mutual information can be negative.
    
    .. math::
        i(x;y) &= h(x) + h(y) - h(x,y) \\\\
               &= \\log\\frac{p(x|y)}{p(x)} \\\\
               &= \\log\\frac{p(y|x)}{p(y)} \\\\
               &= \\log\\frac{p(x,y)}{p(x)p(y)} \\\\

    Parameters
    ----------
    idxs_x : tuple
        The indices of the source variables. Can be multivariate.
    idxs_y : tuple
        The indices of the idxs_y variable. Can be multivariate.
    data : NDArray[np.floating]
        The data in channels x samples format.
    cov : NDArray[np.floating], optional
        The covariance matrix that defines the distribution.
        If none is provided, it is computed from the data object.

    Returns
    -------
     NDArray[np.floating]

    """

    cov_: NDArray[np.floating] = check_cov(cov, data)

    joint: tuple[int, ...] = idxs_x + idxs_y

    h_x: NDArray[np.floating] = local_differential_entropy(
        data = data[idxs_x, :], cov = cov_[np.ix_(idxs_x, idxs_x)]
    )
    h_y: NDArray[np.floating] = local_differential_entropy(
        data = data[idxs_y, :], cov = cov_[np.ix_(idxs_y, idxs_y)]
    )
    h_joint: NDArray[np.floating] = local_differential_entropy(
        data = data[joint, :], cov = cov_[np.ix_(joint, joint)]
    )

    return h_x + h_y - h_joint


def conditional_mutual_information(
    idxs_x: tuple[int, ...],
    idxs_y: tuple[int, ...],
    idxs_z: tuple[int, ...],
    cov: NDArray[np.floating],
) -> float:
    """
    Computes the expected mutual information between a set of variables X
    and Y, conditional on a third set Z.
    
    .. math:: 
        I(X,Y|Z) &= H(X|Z) + H(Y|Z) - H(X,Y|Z) \\\\
                 &= I(X;Y,Z) - I(X;Z)

    Parameters
    ----------
    idxs_x : tuple
        The indices of the X variables. Can be multivariate.
    idxs_y : tuple
        The indices of the Y variable. Can be multivariate.
    idxs_z : tuple
        The indices of the conditioning set. Can be multivariate.
    cov : NDArray[np.floating]
        The covariance matrix that defines the distribution.

    Returns
    -------
    float

    """

    joint: tuple[int, ...] = idxs_x + idxs_y

    h_x_z: float = conditional_entropy(idxs_x, idxs_z, cov)
    h_y_z: float = conditional_entropy(idxs_y, idxs_z, cov)
    h_xy_z: float = conditional_entropy(joint, idxs_z, cov)

    return h_x_z + h_y_z - h_xy_z


def local_conditional_mutual_information(
    idxs_x: tuple[int, ...],
    idxs_y: tuple[int, ...],
    idxs_z: tuple[int, ...],
    data: NDArray[np.floating],
    cov: NDArray[np.floating] = COV_NULL,
):
    """
    Computes the local conditional mutual information between
    two sets of variables X and Y, conditional on another set Z.
    
    .. math:: 
    
        i(x,y|z) &= h(x|z) + h(y|z) - h(x,y|z) \\\\
                 &= i(x;y,z) - i(x;z)

    Parameters
    ----------
    idxs_x : tuple
        The indices of the X variables. Can be multivariate.
    idxs_y : tuple
        The indices of the Y variable. Can be multivariate.
    idxs_z : tuple
        The indices of the conditioning set. Can be multivariate.
    data : NDArray[np.floating]
         The data in channels x samples format.
    cov : NDArray[np.floating], optional
        The covariance matrix that defines the distribution.
        If none is provided, it is computed from the data object.

    Returns
    -------
    NDArray[np.floating]

    """
    cov_: NDArray[np.floating] = check_cov(cov, data)

    joint: tuple[int, ...] = idxs_x + idxs_y

    h_x_z = local_conditional_entropy(idxs_x, idxs_z, data=data, cov=cov_)
    h_y_z = local_conditional_entropy(idxs_y, idxs_z, data=data, cov=cov_)
    h_xy_z = local_conditional_entropy(joint, idxs_z, data=data, cov=cov_)

    return h_x_z + h_y_z - h_xy_z


def kullback_leibler_divergence(
    cov_posterior: NDArray[np.floating], cov_prior: NDArray[np.floating]
) -> float:
    """
    Computes the Gaussian Kullback-Leibler divergence between two :math:`k`-dimensional multivariate Gaussians parameterized by covariance matrices.

    .. math::

        D_{KL}(\\mathcal{N}_0 || \\mathcal{N}_1) = \\frac{1}{2}[ \\operatorname{tr}(\\Sigma_{1}^{-1}\\Sigma_{0}) - k + (\\mu_1 - \\mu_0)^\\mathsf{T} \\Sigma_{1}^{-1}(\\mu_1 - \\mu_0) + \\log\\frac{|\\Sigma_{1}|}{|\\Sigma_{0}|}]

    Parameters
    ----------
    cov_posterior : NDArray[np.floating]
        The covariance matrix that defines the posterior distribution.
    cov_prior : NDArray[np.floating]
        The covariance matrix that defines the prior distribution .

    Returns
    -------
    float

    """
    N: int = cov_prior.shape[0]  # Dimensionality

    assert N == cov_posterior.shape[0], "The covariance matrices must be the same size"

    inv_prior = np.linalg.inv(cov_prior)  # Inverse of Sigma2
    trace_term = np.trace(inv_prior @ cov_posterior)  # tr(Sigma2^{-1} Sigma1)
    log_det_term = (
        np.linalg.slogdet(cov_prior)[1] - np.linalg.slogdet(cov_posterior)[1]
    )  # log(det(Sigma2)/det(Sigma1))

    return 0.5 * (trace_term - N + log_det_term)


def local_kullback_leibler_divergence(
    cov_posterior: NDArray[np.floating],
    cov_prior: NDArray[np.floating],
    data: NDArray[np.floating],
) -> NDArray[np.floating]:
    """
    Computes the local Kullback-Leibler divergence between the
    posterior and the prior for every sample in the data.
    The local KL divergence is a rarely used measure.
    
    .. math::

        d_{kl}^{P||Q}(x) = h^{Q}(x) - h^{P}(x)
    
    Parameters
    ----------
    cov_posterior : NDArray[np.floating]
        The covariance matrix that defines the posterior distribution.
    cov_prior : NDArray[np.floating]
        The covariance matrix that defines the prior distribution .
    data : NDArray[np.floating]
        The data, assumed to be in channels x samples format.

    Returns
    -------
    NDArray[np.floating]
        The local Kullback-Leibler divergence.

    References
    ----------
    Varley, T. F. (2024).
    Generalized decomposition of multivariate information.
    PLOS ONE, 19(2), e0297128.
    https://doi.org/10.1371/journal.pone.0297128

    """

    h_posterior: NDArray[np.floating] = local_differential_entropy(data, cov_posterior)
    h_prior: NDArray[np.floating] = local_differential_entropy(data, cov_prior)

    return h_prior - h_posterior
